from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, initializers, image_data_format
from keras_cv_attention_models.aotnet import AotNet
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    RelativePositionalEmbedding,
    conv2d_no_bias,
    CompatibleExtractPatches,
    make_divisible,
    scaled_dot_product_attention,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "halonet26t": {"imagenet": {256: "6e8848ce6e98a13cd45f65dd68435d00"}},
    "halonet50t": {"imagenet": {256: "6c353144df942cb81a3eb7c140fc9791"}},
    "halonet_se33t": {"imagenet": {256: "7e0afb7f8fb6459491b8a46ad80bcd91"}},
    "halonext_eca26t": {"imagenet": {256: "630037a5c135bceacd0691a22855eb7e"}},
    "haloregnetz_b": {"imagenet": {224: "e889647682d1c554de032d376acf0c48"}},
    "halobotnet50t": {"imagenet": {256: "0af1faad1a81e468d6e670e9fc253edc"}},
}


def halo_attention(
    inputs, num_heads=8, key_dim=0, block_size=4, halo_size=1, strides=1, out_shape=None, out_weight=True, out_bias=False, attn_dropout=0, name=None
):
    height_axis, width_axis, channel_axis = (1, 2, 3) if image_data_format() == "channels_last" else (2, 3, 1)
    filters = inputs.shape[channel_axis]
    if key_dim > 1:
        key_dim = key_dim  # Specified one
    elif key_dim > 0:
        key_dim = make_divisible(filters * key_dim, divisor=8) // num_heads  # regard as key_dim_ratio
    else:
        key_dim = filters // num_heads  # Default value
    # qk_scale = float(1.0 / tf.math.sqrt(tf.cast(key_dim, "float32")))
    qk_scale = 1.0 / (float(key_dim) ** 0.5)
    out_shape = filters if out_shape is None else out_shape
    emb_dim = num_heads * key_dim
    kv_kernel = block_size + halo_size * 2
    if block_size % strides != 0:
        strides = 1
        avg_pool_down = True
    else:
        avg_pool_down = False
    query_block = block_size // strides

    query = conv2d_no_bias(inputs, emb_dim, kernel_size=1, strides=strides, name=name and name + "query_")
    # print(f">>>> {inputs.shape = }, {query.shape = }, {block_size = }, {strides = }")
    # attn_query = rearrange(query, "B (h hb) (w wb) (hd c) -> B hd h w (hb wb) c", hb=query_block, wb=query_block, hd=num_heads)
    # pos_query = rearrange(attn_query, "B hd h w (hb wb) c -> B (hd h w) hb wb c", hb=query_block, wb=query_block)
    query = query if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(query)  # channels_first -> channels_last
    hh_qq, ww_qq, cc_qq = query.shape[1] // query_block, query.shape[2] // query_block, emb_dim // num_heads
    query = functional.reshape(query, [-1, hh_qq, query_block, ww_qq, query_block, num_heads, cc_qq])
    query = functional.transpose(query, [0, 5, 1, 3, 2, 4, 6])  # [batch, num_heads, hh, ww, query_block, query_block, key_dim]
    # attn_query = [batch, num_heads, hh, ww, query_block * query_block, key_dim]
    attn_query = functional.reshape(query, [-1, num_heads, hh_qq, ww_qq, query_block * query_block, cc_qq]) * qk_scale  # qk_scale NOT multiplied with pos_query
    # pos_query = [batch, num_heads * hh * ww, query_block, query_block, key_dim]
    pos_query = functional.reshape(query, [-1, num_heads * hh_qq * ww_qq, query_block, query_block, cc_qq])

    # key_value = [batch, height, width, key_dim + out_shape]
    key_value = conv2d_no_bias(inputs, emb_dim + out_shape, kernel_size=1, use_bias=False, name=name and name + "key_value_")
    key_value = key_value if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(key_value)  # channels_first -> channels_last
    kv_padded = functional.pad(key_value, [[0, 0], [halo_size, halo_size], [halo_size, halo_size], [0, 0]])
    # kv_inp = [batch, hh, ww, kv_kernel * kv_kernel * (key_dim + out_shape)]
    if backend.is_torch_backend:
        kv_inp = functional.extract_patches(kv_padded, sizes=kv_kernel, strides=block_size, padding="valid")
    else:
        # kv_inp = tf.image.extract_patches(kv_padded, sizes=sizes, strides=strides, rates=[1, 1, 1, 1], padding="valid")
        kv_inp = CompatibleExtractPatches(sizes=kv_kernel, strides=block_size, padding="valid")(kv_padded)
    # kv_inp = rearrange(kv_inp, "B h w (hb wb hd c) -> B hd h w (hb wb) c", hb=kv_kernel, wb=kv_kernel, hd=num_heads)
    _, hh_kk, ww_kk, cc = kv_inp.shape
    cc_kk = cc // num_heads // kv_kernel // kv_kernel
    kv_inp = functional.reshape(kv_inp, [-1, hh_kk, ww_kk, kv_kernel, kv_kernel, num_heads, cc_kk])
    kv_inp = functional.transpose(kv_inp, [0, 5, 1, 2, 3, 4, 6])
    kv_inp = functional.reshape(kv_inp, [-1, num_heads, hh_kk, ww_kk, kv_kernel * kv_kernel, cc_kk])

    # key = [batch, num_heads, hh, ww, kv_kernel * kv_kernel, key_dim]
    # value = [batch, num_heads, hh, ww, kv_kernel * kv_kernel, out_dim]
    key, value = functional.split(kv_inp, [emb_dim // num_heads, out_shape // num_heads], axis=-1)

    # scaled_dot_product_attention
    # print(f">>>> {attn_query.shape = }, {key.shape = }, {value.shape = }, {kv_inp.shape = }, {pos_query.shape = }, {num_heads = }")
    # attention_scores = [batch, num_heads, hh, ww, query_block * query_block, kv_kernel * kv_kernel]
    # attention_scores = layers.Lambda(lambda xx: functional.matmul(xx[0], xx[1], transpose_b=True))([attn_query, key])
    attention_scores = attn_query @ functional.transpose(key, [0, 1, 2, 3, 5, 4])
    # pos = [batch, num_heads * hh * ww, query_block, query_block, kv_kernel, kv_kernel]
    pos = RelativePositionalEmbedding(position_height=kv_kernel, name=name and name + "pos_emb")(pos_query)
    # print(f">>>> {pos.shape = }, {attention_scores.shape = }")
    pos = functional.reshape(pos, [-1, *attention_scores.shape[1:]])
    attention_scores = layers.Add()([attention_scores, pos])
    # attention_scores = tf.nn.softmax(attention_scores, axis=-1)
    attention_scores = layers.Softmax(axis=-1, name=name and name + "attention_scores")(attention_scores)

    if attn_dropout > 0:
        attention_scores = layers.Dropout(attn_dropout, name=name and name + "attn_drop")(attention_scores)

    # attention_output = [batch, num_heads, hh, ww, query_block * query_block, out_dim]
    # attention_output = layers.Lambda(lambda xx: functional.matmul(xx[0], xx[1]))([attention_scores, value])
    attention_output = attention_scores @ value
    # attention_output = rearrange(attention_output, "B hd h w (hb wb) c -> B (h hb) (w wb) (hd c)", hb=query_block, wb=query_block)
    _, heads, hh_aa, ww_aa, patch, cc_aa = attention_output.shape
    attention_output = functional.reshape(attention_output, [-1, heads, hh_aa, ww_aa, query_block, query_block, cc_aa])
    attention_output = functional.transpose(attention_output, [0, 2, 4, 3, 5, 1, 6])
    attention_output = functional.reshape(attention_output, [-1, hh_aa * query_block, ww_aa * query_block, heads * cc_aa])
    # print(f">>>> {attention_output.shape = }, {attention_scores.shape = }")

    cur_data_format = "channels_last"
    if avg_pool_down:
        attention_output = attention_output if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(attention_output)
        attention_output = layers.AvgPool2D(2, strides=2, name=name and name + "avg_pool")(attention_output)
        cur_data_format = image_data_format()
    if out_weight:
        # [batch, hh, ww, num_heads * out_dim] * [out, out] --> [batch, hh, ww, out]
        attention_output = attention_output if cur_data_format == "channels_last" else layers.Permute([2, 3, 1])(attention_output)
        attention_output = layers.Dense(out_shape, use_bias=out_bias, name=name and name + "output")(attention_output)
        cur_data_format = "channels_last"
    return attention_output if cur_data_format == image_data_format() else layers.Permute([3, 1, 2])(attention_output)


BLOCK_CONFIGS = {
    "h0": {  # rv = 1, rb = 0.5
        "halo_block_size": 8,  # b
        "halo_halo_size": 3,  # h
        "halo_expansion": 1,  # rv
        "expansion": 0.5,  # rb
        "output_num_features": -1,  # df
        "num_blocks": [3, 3, 7, 3],
        "num_heads": [4, 8, 8, 8],
    },
    "h1": {  # rv = 1, rb = 1
        "halo_block_size": 8,
        "halo_halo_size": 3,
        "halo_expansion": 1,  # rv
        "expansion": 1,  # rb
        "output_num_features": -1,  # df
        "num_blocks": [3, 3, 10, 3],
        "num_heads": [4, 8, 8, 8],
    },
    "h2": {  # rv = 1, rb = 1.25
        "halo_block_size": 8,
        "halo_halo_size": 3,
        "halo_expansion": 1,  # rv
        "expansion": 1.25,  # rb
        "output_num_features": -1,  # df
        "num_blocks": [3, 3, 11, 3],
        "num_heads": [4, 8, 8, 8],
    },
    "h3": {  # rv = 1, rb = 1.5
        "halo_block_size": 10,
        "halo_halo_size": 3,
        "halo_expansion": 1,  # rv
        "expansion": 1.5,  # rb
        "output_num_features": 1024,  # df
        "num_blocks": [3, 3, 12, 3],
        "num_heads": [4, 8, 8, 8],
    },
    "h4": {  # rv = 1, rb = 3
        "halo_block_size": 12,
        "halo_halo_size": 2,
        "halo_expansion": 1,  # rv
        "expansion": 3,  # rb
        "output_num_features": 1280,  # df
        "num_blocks": [3, 3, 12, 3],
        "num_heads": [4, 8, 8, 8],
    },
    "h5": {  # rv = 2.5, rb = 2
        "halo_block_size": 14,
        "halo_halo_size": 2,
        "halo_expansion": 2.5,  # rv
        "expansion": 2,  # rb
        "output_num_features": 1536,  # df
        "num_blocks": [3, 3, 23, 3],
        "num_heads": [4, 8, 8, 8],
    },
    "h6": {  # rv = 3, rb = 2.75
        "halo_block_size": 8,
        "halo_halo_size": 4,
        "halo_expansion": 3,  # rv
        "expansion": 2.75,  # rb
        "output_num_features": 1536,  # df
        "num_blocks": [3, 3, 24, 3],
        "num_heads": [4, 8, 8, 8],
    },
    "h7": {  # rv = 4, rb = 3.5
        "halo_block_size": 10,
        "halo_halo_size": 3,
        "halo_expansion": 4,  # rv
        "num_heads": [4, 8, 8, 8],
        "expansion": 3.5,  # rb
        "output_num_features": 2048,  # df
        "num_blocks": [3, 3, 26, 3],
    },
}


def HaloNet(
    input_shape=(256, 256, 3),
    activation="swish",
    expansion=4,
    halo_block_size=4,
    halo_halo_size=1,
    halo_expansion=1,
    num_heads=8,
    pretrained=None,
    request_resolution=256,
    **kwargs
):
    attn_types = "halo"
    if isinstance(num_heads, (list, tuple)):
        attn_params = [
            {"block_size": halo_block_size, "halo_size": halo_halo_size, "halo_expansion": halo_expansion, "num_heads": hh, "out_weight": False}
            for hh in num_heads
        ]
    else:
        attn_params = {
            "block_size": halo_block_size,
            "halo_size": halo_halo_size,
            "halo_expansion": halo_expansion,
            "num_heads": num_heads,
            "out_weight": False,
        }
    out_channels = [ii * expansion for ii in [64, 128, 256, 512]]
    hidden_channel_ratio = 1 / expansion
    model = AotNet(
        input_shape=input_shape,
        out_channels=out_channels,
        hidden_channel_ratio=hidden_channel_ratio,
        activation=activation,
        attn_types=attn_types,
        attn_params=attn_params,
        **kwargs
    )
    reload_model_weights(model, PRETRAINED_DICT, "halonet", pretrained, RelativePositionalEmbedding)
    return model


def HaloNetH0(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return HaloNet(**BLOCK_CONFIGS["h0"], model_name="haloneth0", request_resolution=256, **locals(), **kwargs)


def HaloNetH1(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return HaloNet(**BLOCK_CONFIGS["h1"], model_name="haloneth1", request_resolution=256, **locals(), **kwargs)


def HaloNetH2(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return HaloNet(**BLOCK_CONFIGS["h2"], model_name="haloneth2", request_resolution=256, **locals(), **kwargs)


def HaloNetH3(input_shape=(320, 320, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return HaloNet(**BLOCK_CONFIGS["h3"], model_name="haloneth3", request_resolution=320, **locals(), **kwargs)


def HaloNetH4(input_shape=(384, 384, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return HaloNet(**BLOCK_CONFIGS["h4"], model_name="haloneth4", request_resolution=384, **locals(), **kwargs)


def HaloNetH5(input_shape=(448, 448, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return HaloNet(**BLOCK_CONFIGS["h5"], model_name="haloneth5", request_resolution=448, **locals(), **kwargs)


def HaloNetH6(input_shape=(512, 512, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return HaloNet(**BLOCK_CONFIGS["h6"], model_name="haloneth6", request_resolution=512, **locals(), **kwargs)


def HaloNetH7(input_shape=(600, 600, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    # input_shape should be divisible by `int(tf.reduce_prod(strides) * halo_block_size)`, may using 640 here
    return HaloNet(**BLOCK_CONFIGS["h7"], model_name="haloneth7", request_resolution=600, **locals(), **kwargs)


@register_model
def HaloNet26T(input_shape=(256, 256, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 2, 2]
    attn_types = [None, None, [None, "halo"], "halo"]
    attn_params = [
        None,
        None,
        [None, {"block_size": 8, "halo_size": 2, "num_heads": 8, "out_weight": False}],
        {"block_size": 8, "halo_size": 2, "num_heads": 8, "out_weight": False},
    ]
    # key_dim = 16
    stem_type = "tiered"
    model = AotNet(model_name="halonet26t", **locals(), **kwargs)
    reload_model_weights(model, PRETRAINED_DICT, "halonet", pretrained, RelativePositionalEmbedding)
    return model


@register_model
def HaloNet50T(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 6, 3]
    attn_types = [None, [None, None, None, "halo"], [None, "halo"] * 3, [None, "halo", None]]
    attn_params = [
        None,
        [None, None, None, {"block_size": 8, "halo_size": 3, "num_heads": 4, "out_weight": False}],
        [None, {"block_size": 8, "halo_size": 3, "num_heads": 8, "out_weight": False}] * 3,
        [None, {"block_size": 8, "halo_size": 3, "num_heads": 8, "out_weight": False}, None],
    ]
    # key_dim = 16
    stem_type = "tiered"
    model = AotNet(model_name="halonet50t", **locals(), **kwargs)
    reload_model_weights(model, PRETRAINED_DICT, "halonet", pretrained, RelativePositionalEmbedding)
    return model


@register_model
def HaloNetSE33T(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 3, 3, 2]
    attn_types = [None, [None, None, "halo"], [None, None, "halo"], "halo"]
    attn_params = [
        None,
        [None, None, {"block_size": 8, "halo_size": 3, "num_heads": 8, "out_weight": False}],
        [None, None, {"block_size": 8, "halo_size": 3, "num_heads": 8, "out_weight": False}],
        {"block_size": 8, "halo_size": 3, "num_heads": 8, "out_weight": False},
    ]
    se_ratio = 1 / 16
    out_channels = [256, 512, 1024, 1536]
    hidden_channel_ratio = [1 / 4, 1 / 4, 1 / 4, 1 / 3]
    # key_dim = 16
    stem_type = "tiered"
    stem_last_strides = 2
    stem_downsample = False
    output_num_features = 1280
    model = AotNet(model_name="halonet_se33t", **locals(), **kwargs)
    reload_model_weights(model, PRETRAINED_DICT, "halonet", pretrained, RelativePositionalEmbedding)
    return model


@register_model
def HaloNextECA26T(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 2, 2]
    attn_types = [None, None, [None, "halo"], "halo"]
    attn_params = [
        None,
        None,
        [None, {"block_size": 8, "halo_size": 2, "num_heads": 8, "key_dim": 16, "out_weight": False}],
        {"block_size": 8, "halo_size": 2, "num_heads": 8, "key_dim": 16, "out_weight": False},
    ]
    use_eca = True
    groups = [4, 8, 16, 32]
    stem_type = "tiered"
    model = AotNet(model_name="halonext_eca26t", **locals(), **kwargs)
    reload_model_weights(model, PRETRAINED_DICT, "halonet", pretrained, RelativePositionalEmbedding)
    return model


@register_model
def HaloRegNetZB(input_shape=(224, 224, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 6, 12, 2]
    strides = [2, 2, 2, 2]
    out_channels = [48, 96, 192, 288]
    # timm bottle_in=True mode, the first ratio in each stack is `ratio * previous_channel`
    hidden_channel_ratio = [[32 * 3 / 48, 3], [1.5] + [3] * 5, [1.5] + [3] * 11, [192 * 3 / 288, 3]]
    use_block_output_activation = False  # timm linear_out=True mode
    stem_type = "kernel_3x3"
    stem_width = 32
    stem_downsample = False
    attn_types = [None, None, [None, None, None, "halo"] * 3, "halo"]
    # "activation": "relu" is for se module.
    attn_params = {"block_size": 7, "halo_size": 2, "num_heads": 8, "key_dim": 0.33, "out_weight": False, "activation": "relu"}
    se_ratio = 0.25
    group_size = 16
    shortcut_type = None
    output_num_features = 1536
    model = AotNet(model_name="haloregnetz_b", **locals(), **kwargs)
    reload_model_weights(model, PRETRAINED_DICT, "halonet", pretrained, RelativePositionalEmbedding)
    return model


@register_model
def HaloBotNet50T(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 6, 3]
    attn_types = [None, [None, "halo"] * 2, [None, "halo"] * 3, [None, "bot", None]]
    attn_params = [
        None,
        [None, {"block_size": 8, "halo_size": 3, "num_heads": 8, "out_weight": False}] * 2,
        [None, {"block_size": 8, "halo_size": 3, "num_heads": 8, "out_weight": False}] * 3,
        [None, {"num_heads": 4, "out_weight": False}, None],
    ]
    # key_dim = 16
    stem_type = "tiered"
    stem_last_strides = 2
    stem_downsample = False
    model = AotNet(model_name="halobotnet50t", **locals(), **kwargs)
    reload_model_weights(model, PRETRAINED_DICT, "halonet", pretrained, RelativePositionalEmbedding)
    return model



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    add_with_layer_scale_and_drop_block,
    batchnorm_with_activation,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    layer_norm,
    mlp_block,
    multi_head_self_attention,
    output_block,
    PositionalEmbedding,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

LAYER_NORM_EPSILON = 1e-6

PRETRAINED_DICT = {
    "iformer_base": {"imagenet": {224: "90c8fe1307bc8cbd73f5a7358e65956d", 384: "11d150c128a65f8dcc95871e87209491"}},
    "iformer_large": {"imagenet": {224: "2a116d5780a551846761aa43b1d74395", 384: "b21ae4484d1f5da7f1425ac301248e29"}},
    "iformer_small": {"imagenet": {224: "de2f8262da0cc1c6df43b32b827444f1", 384: "e1c7f4e52abf3514437138c62194d2d7"}},
}


def attention_low_frequency_mixer(inputs, num_heads=4, pool_size=1, dropout=0, name=""):
    height_axis, width_axis, channel_axis = (1, 2, 3) if image_data_format() == "channels_last" else (2, 3, 1)
    # print(f"{inputs.shape = }, {pool_size = }")
    if pool_size > 1:
        orign_height, orign_width = inputs.shape[height_axis], inputs.shape[width_axis]
        inputs = layers.AvgPool2D(pool_size, strides=pool_size, padding="same", name=name + "avg_down")(inputs)
    nn = inputs if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(inputs)  # channels_first -> channels_last
    nn = multi_head_self_attention(nn, num_heads=num_heads, qkv_bias=True, out_weight=False, attn_dropout=dropout, name=name + "attn_")
    nn = nn if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(nn)  # channels_last -> channels_first

    if pool_size > 1:
        nn = layers.UpSampling2D(size=pool_size, name=name + "up")(nn)
        if nn.shape[height_axis] != orign_height or nn.shape[width_axis] != orign_width:
            nn = nn[:, :orign_height, :orign_width] if image_data_format() == "channels_last" else nn[:, :, :orign_height, :orign_width]
    return nn


def conv_high_frequency_mixer(inputs, activation="gelu", name=""):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    nn = conv2d_no_bias(inputs, inputs.shape[channel_axis] * 2, kernel_size=1, name=name)
    nn = depthwise_conv2d_no_bias(nn, kernel_size=3, padding="same", name=name)
    nn = activation_by_name(nn, activation=activation, name=name)
    return nn


def pool_high_frequency_mixer(inputs, activation="gelu", name=""):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    nn = layers.MaxPool2D(3, strides=1, padding="same", name=name + "max")(inputs)
    nn = conv2d_no_bias(nn, inputs.shape[channel_axis] * 2, kernel_size=1, use_bias=True, name=name)
    nn = activation_by_name(nn, activation=activation, name=name)
    return nn


def conv_pool_attention_mixer(inputs, num_heads=4, key_dim=0, num_attn_low_heads=1, pool_size=1, dropout=0, activation="gelu", name=""):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    input_channel = inputs.shape[channel_axis]
    key_dim = key_dim if key_dim > 0 else input_channel // num_heads
    attention_channels = num_attn_low_heads * key_dim
    conv_channels = (input_channel - attention_channels) // 2
    pool_channels = input_channel - attention_channels - conv_channels
    conv_branch, pool_branch, attention_branch = functional.split(inputs, [conv_channels, pool_channels, attention_channels], axis=channel_axis)
    # print(f"{key_dim = }, {num_heads = }, {num_attn_low_heads = }, {attention_channels = }")

    conv_branch = conv_high_frequency_mixer(conv_branch, activation=activation, name=name + "high_conv_branch_")
    pool_branch = pool_high_frequency_mixer(pool_branch, activation=activation, name=name + "high_pool_branch_")
    attention_branch = attention_low_frequency_mixer(attention_branch, num_heads=num_attn_low_heads, pool_size=pool_size, dropout=dropout, name=name + "low_")
    high_low = functional.concat([conv_branch, pool_branch, attention_branch], axis=channel_axis)
    # print(f"{conv_branch.shape = }, {pool_branch.shape = }, {attention_branch.shape = }, {high_low.shape = }")

    high_low_fused = depthwise_conv2d_no_bias(high_low, kernel_size=3, padding="same", name=name + "fuse_")
    high_low_out = layers.Add()([high_low, high_low_fused])

    out = conv2d_no_bias(high_low_out, input_channel, kernel_size=1, use_bias=True, name=name + "output_")
    if dropout > 0:
        out = layers.Dropout(dropout)(out)
    return out


def attention_mlp_block(inputs, num_heads=8, num_attn_low_heads=1, pool_size=1, mlp_ratio=4, layer_scale=0.1, drop_rate=0, activation="gelu", name=""):
    input_channel = inputs.shape[-1]  # Channels_last only

    """ attention """
    nn = layer_norm(inputs, epsilon=LAYER_NORM_EPSILON, axis=-1, name=name + "attn_")
    nn = nn if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(nn)  # channels_last -> channels_first
    nn = conv_pool_attention_mixer(nn, num_heads, num_attn_low_heads=num_attn_low_heads, pool_size=pool_size, activation=activation, name=name + "attn_")
    nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(nn)  # channels_first -> channels_last
    attn_out = add_with_layer_scale_and_drop_block(inputs, nn, layer_scale=layer_scale, drop_rate=drop_rate, axis=-1, name=name + "attn_")

    """ MLP """
    nn = layer_norm(attn_out, epsilon=LAYER_NORM_EPSILON, axis=-1, name=name + "mlp_")
    nn = mlp_block(nn, input_channel * mlp_ratio, activation=activation, name=name + "mlp_")
    nn = add_with_layer_scale_and_drop_block(attn_out, nn, layer_scale=layer_scale, drop_rate=drop_rate, axis=-1, name=name + "mlp_")
    return nn


def InceptionTransformer(
    num_blocks=[3, 3, 9, 3],
    embed_dims=[96, 192, 320, 384],
    num_heads=[3, 6, 10, 12],
    num_attn_low_heads=[1, 3, [7] * 4 + [9] * 5, 11],
    pool_sizes=[2, 2, 1, 1],
    mlp_ratios=4,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    dropout=0,
    layer_scales=[0, 0, 1e-6, 1e-6],
    classifier_activation="softmax",
    pretrained=None,
    model_name="iformer",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)

    """ Stem """
    nn = conv2d_no_bias(inputs, embed_dims[0] // 2, kernel_size=3, strides=2, padding="same", use_bias=True, name="stem_1_")
    nn = batchnorm_with_activation(nn, activation=activation, name="stem_1_")
    nn = conv2d_no_bias(nn, embed_dims[0], kernel_size=3, strides=2, padding="same", use_bias=True, name="stem_2_")
    nn = batchnorm_with_activation(nn, activation=None, name="stem_2_")

    """ stacks """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, embed_dim) in enumerate(zip(num_blocks, embed_dims)):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            nn = conv2d_no_bias(nn, embed_dim, 3, strides=2, padding="same", use_bias=True, name=stack_name + "downsample_")
            nn = batchnorm_with_activation(nn, activation=None, name=stack_name + "downsample_")  # Using epsilon=1e-5
        nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(nn)  # channels_first -> channels_last
        nn = PositionalEmbedding(name=stack_name + "positional_embedding")(nn)

        stack_num_attn_low_head = num_attn_low_heads[stack_id] if isinstance(num_attn_low_heads, (list, tuple)) else num_attn_low_heads
        num_head = num_heads[stack_id] if isinstance(num_heads, (list, tuple)) else num_heads
        mlp_ratio = mlp_ratios[stack_id] if isinstance(mlp_ratios, (list, tuple)) else mlp_ratios
        pool_size = pool_sizes[stack_id] if isinstance(pool_sizes, (list, tuple)) else pool_sizes
        layer_scale = layer_scales[stack_id] if isinstance(layer_scales, (list, tuple)) else layer_scales
        for block_id in range(num_block):
            name = stack_name + "block{}_".format(block_id + 1)
            num_attn_low_head = stack_num_attn_low_head[block_id] if isinstance(stack_num_attn_low_head, (list, tuple)) else stack_num_attn_low_head
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            nn = attention_mlp_block(nn, num_head, num_attn_low_head, pool_size, mlp_ratio, layer_scale, block_drop_rate, activation=activation, name=name)
            global_block_id += 1
        nn = nn if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(nn)  # channels_last -> channels_first
    nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, name="pre_output_")

    nn = output_block(nn, num_classes=num_classes, drop_rate=dropout, classifier_activation=classifier_activation)
    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "iformer", pretrained, PositionalEmbedding)
    return model


@register_model
def IFormerSmall(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return InceptionTransformer(**locals(), model_name="iformer_small", **kwargs)


@register_model
def IFormerBase(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [4, 6, 14, 6]
    embed_dims = [96, 192, 384, 512]
    num_heads = [3, 6, 12, 16]
    num_attn_low_heads = [1, 3, [8] * 7 + [10] * 7, 15]
    return InceptionTransformer(**locals(), model_name="iformer_base", **kwargs)


@register_model
def IFormerLarge(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [4, 6, 18, 8]
    embed_dims = [96, 192, 448, 640]
    num_heads = [3, 6, 14, 20]
    num_attn_low_heads = [1, 3, [10] * 9 + [12] * 9, 19]
    return InceptionTransformer(**locals(), model_name="iformer_large", **kwargs)



from keras_cv_attention_models.uniformer.uniformer import (
    Uniformer,
    UniformerSmall32,
    UniformerSmall64,
    UniformerSmallPlus32,
    UniformerSmallPlus64,
    UniformerBase32,
    UniformerBase64,
    UniformerLarge64,
    multi_head_self_attention,
)

__head_doc__ = """
Keras implementation of [UniFormer](https://github.com/Sense-X/UniFormer/tree/main/image_classification).
Paper [PDF 2201.09450 UniFormer: Unifying Convolution and Self-attention for Visual Recognition](https://arxiv.org/pdf/2201.09450.pdf).
"""

__tail_doc__ = """  block_types: block types for each stack,
      - `conv` or any `c` / `C` starts word, means `res_MBConv` block.
      - `transfrom` or any not `c` / `C` starts word, means `res_mhsa` + `res_ffn` block.
      value could be in format like `"cctt"` or `"CCTT"` or `["conv", "conv", "transfrom", "transform"]`.
  stem_width: output dimension for stem block. Default -1 means using `out_channels[0]`
  qkv_bias: boolean value if useing bias for `qkv` in `mhsa` block, default `True`.
  mlp_ratio: channel expansion ratio for mlp hidden layers, default 4.
  layer_scale: layer scale init value. Default `-1` means not applying, any value `>=0` will add a scale value for each block output.
      [Going deeper with Image Transformers](https://arxiv.org/pdf/2103.17239.pdf).
  mix_token: boolean value if using mix token augment, should work together with `token_label_top`, default `False`.
  token_label_top: boolean value if output `class + token header`, Default `False`.
      [All Tokens Matter: Token Labeling for Training Better Vision Transformers](https://arxiv.org/pdf/2104.10858.pdf).
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `relu`.
  mlp_drop_rate: dropout rate for mlp blocks.
  attn_drop_rate: dropout rate for mhsa attention scores.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: dropout rate if top layers is included.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  pretrained: None or one of ["imagenet", "token_label"].

Returns:
    A `keras.Model` instance.
"""

Uniformer.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: output channels for each stack.
  head_dimension: heads number for transformer block. `32` for `xxx32` models and `64` for `xxx64` models.
  use_conv_stem: boolean value if using `Conv` stem or `Patch` stem. `True` for `UniformerSmallPlus*`, `False` for others.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model                 | Params | FLOPs  | Input | Top1 Acc |
  | --------------------- | ------ | ------ | ----- | -------- |
  | UniformerSmall32 + TL | 22M    | 3.66G  | 224   | 83.4     |
  | UniformerSmall64      | 22M    | 3.66G  | 224   | 82.9     |
  | - Token Labeling      | 22M    | 3.66G  | 224   | 83.4     |
  | UniformerSmallPlus32  | 24M    | 4.24G  | 224   | 83.4     |
  | - Token Labeling      | 24M    | 4.24G  | 224   | 83.9     |
  | UniformerSmallPlus64  | 24M    | 4.23G  | 224   | 83.4     |
  | - Token Labeling      | 24M    | 4.23G  | 224   | 83.6     |
  | UniformerBase32 + TL  | 50M    | 8.32G  | 224   | 85.1     |
  | UniformerBase64       | 50M    | 8.31G  | 224   | 83.8     |
  | - Token Labeling      | 50M    | 8.31G  | 224   | 84.8     |
  | UniformerLarge64 + TL | 100M   | 19.79G | 224   | 85.6     |
  | UniformerLarge64 + TL | 100M   | 63.11G | 384   | 86.3     |
"""

UniformerSmall32.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

UniformerSmall64.__doc__ = UniformerSmall32.__doc__
UniformerSmallPlus32.__doc__ = UniformerSmall32.__doc__
UniformerSmallPlus64.__doc__ = UniformerSmall32.__doc__
UniformerBase32.__doc__ = UniformerSmall32.__doc__
UniformerBase64.__doc__ = UniformerSmall32.__doc__
UniformerLarge64.__doc__ = UniformerSmall32.__doc__

multi_head_self_attention.__doc__ = """
Typical multi head self attention block, should work similar with `keras.layers.MultiHeadAttention`. Defined as function, not layer.

Args:
  inputs: input tensor.
  num_heads: Number of attention heads.
  key_dim: Size of each attention head for query and key. Default `0` for `key_dim = inputs.shape[-1] // num_heads`.
  out_shape: The expected shape of an output tensor. If not specified, projects back to the input dim.
  out_weight: Boolean, whether use an ouput dense.
  qkv_bias: Boolean, whether the qkv dense layer use bias vectors/matrices.
  out_bias: Boolean, whether the ouput dense layer use bias vectors/matrices.
  attn_dropout: Dropout probability for attention scores.
  output_dropout: Dropout probability for attention output.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 16, 256])
>>> nn = attention_layers.multi_head_self_attention(inputs, num_heads=4, out_shape=512)
>>> print(f"{nn.shape = }")
# nn.shape = TensorShape([None, 14, 16, 512])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
>>> print({ii.name: ii.shape for ii in mm.weights})
# {'dense/kernel:0': TensorShape([256, 1024]), 'dense_1/kernel:0': TensorShape([512, 512])}
"""



import numpy as np
from tqdm.auto import tqdm
from PIL import Image
import keras_cv_attention_models
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, models, initializers, image_data_format
from keras_cv_attention_models.clip import tokenizer
from keras_cv_attention_models.models import register_model, FakeModelWrapper


@register_model
class StableDiffusion(FakeModelWrapper):  # FakeModelWrapper providing save / load / cuda class methods
    def __init__(
        self,
        clip_model="beit.ViTTextLargePatch14",
        unet_model="stable_diffusion.UNet",
        decoder_model="stable_diffusion.Decoder",
        encoder_model=None,  # or "stable_diffusion.Encoder"
        clip_model_kwargs={},
        unet_model_kwargs={},
        decoder_model_kwargs={},
        encoder_model_kwargs={},
        caption_tokenizer="SimpleTokenizer",
        pretrained="default",  # default for using model preset weights, set None to disable all pretrained, for specific model, set related xxx_model_kwargs
        num_steps=50,
        num_training_steps=1000,
        ddim_discretize="uniform",
        linear_start=0.00085,
        linear_end=0.0120,
        ddim_eta=0.0,
        name="stable_diffusion",
    ):
        if pretrained != "default":
            clip_model_kwargs.update({"pretrained": pretrained})
            unet_model_kwargs.update({"pretrained": pretrained})
            decoder_model_kwargs.update({"pretrained": pretrained})
            encoder_model_kwargs.update({"pretrained": pretrained})

        """ Dynamic input shape """
        if image_data_format() == "channels_last":
            image_shape, latents_input_shape = [None, None, 3], [None, None, 4]  # [512, 512, 3], [64, 64, 4]
        else:
            image_shape, latents_input_shape = [3, None, None], [4, None, None]  # [3, 512, 512], [4, 64, 64]
        self.image_shape = (None, *image_shape)
        self.latents_input_shape = (None, *latents_input_shape)

        """ tokenizer """
        if isinstance(caption_tokenizer, str) and hasattr(tokenizer, caption_tokenizer):
            self.caption_tokenizer = getattr(tokenizer, caption_tokenizer)()
        elif isinstance(caption_tokenizer, str):  # tiktoken one
            self.caption_tokenizer = tokenizer.TikToken(caption_tokenizer)
        else:
            self.caption_tokenizer = tokenizer

        """ clip_model """
        clip_model_kwargs.update({"vocab_size": self.caption_tokenizer.vocab_size, "include_top": False})
        self.clip_model = self.build_model(clip_model, **clip_model_kwargs)
        self.uncond_prompt = None  # Init later after load weights

        """ encoder, could be None """
        encoder_model_kwargs["input_shape"] = encoder_model_kwargs.pop("input_shape", image_shape)
        self.encoder_model = self.build_model(encoder_model, **encoder_model_kwargs)
        self.encoder_model_kwargs = encoder_model_kwargs  # In case need this later, when calling image_to_image

        """ unet """
        unet_model_kwargs["input_shape"] = unet_model_kwargs.pop("input_shape", latents_input_shape)
        self.unet_model = self.build_model(unet_model, **unet_model_kwargs)

        """ decoder """
        decoder_model_kwargs["input_shape"] = decoder_model_kwargs.pop("input_shape", latents_input_shape)
        self.decoder_model = self.build_model(decoder_model, **decoder_model_kwargs)
        self.output_shape = self.decoder_model.output_shape

        """ Gather models and sampler """
        self.models = [self.clip_model, self.unet_model, self.decoder_model] + ([] if self.encoder_model is None else [self.encoder_model])
        self.num_steps, self.num_training_steps, self.ddim_discretize, self.ddim_eta = num_steps, num_training_steps, ddim_discretize, ddim_eta
        self.linear_start, self.linear_end = linear_start, linear_end
        self.init_ddim_sampler(num_steps, num_training_steps, ddim_discretize, linear_start, linear_end, ddim_eta)
        super().__init__(self.models, name=name)

    def build_model(self, model, **model_kwargs):
        if not isinstance(model, str) or len(model) == 0:
            return model

        model_split = model.split(".")
        if len(model_split) == 2:
            model_class = getattr(getattr(keras_cv_attention_models, model_split[0]), model_split[1])
        else:
            model_class = getattr(keras_cv_attention_models.models, model_split[0])
        return model_class(**model_kwargs)

    def init_ddim_sampler(self, num_steps=50, num_training_steps=1000, ddim_discretize="uniform", linear_start=0.00085, linear_end=0.0120, ddim_eta=0.0):
        # DDIM sampling from the paper [Denoising Diffusion Implicit Models](https://papers.labml.ai/paper/2010.02502)
        # num_steps, num_training_steps, ddim_discretize, linear_start, linear_end, ddim_eta = 50, 1000, "uniform", 0.00085, 0.0120, 0
        if ddim_discretize == "quad":
            time_steps = ((np.linspace(0, np.sqrt(num_training_steps * 0.8), num_steps)) ** 2).astype(int) + 1
        else:  # "uniform"
            interval = num_training_steps // num_steps
            time_steps = np.arange(0, num_training_steps, interval) + 1

        beta = np.linspace(linear_start**0.5, linear_end**0.5, num_training_steps, dtype="float64") ** 2
        alpha_bar = np.cumprod(1.0 - beta, axis=0).astype("float32")

        ddim_alpha = alpha_bar[time_steps]
        ddim_alpha_prev = np.concatenate([alpha_bar[:1], alpha_bar[time_steps[:-1]]])
        ddim_sigma = ddim_eta * ((1 - ddim_alpha_prev) / (1 - ddim_alpha) * (1 - ddim_alpha / ddim_alpha_prev)) ** 0.5

        self.time_steps, self.ddim_alpha, self.ddim_alpha_prev, self.ddim_sigma = time_steps, ddim_alpha, ddim_alpha_prev, ddim_sigma
        self.ddim_alpha_sqrt, self.ddim_sqrt_one_minus_alpha = ddim_alpha**0.5, (1.0 - ddim_alpha) ** 0.5

    def gaussian_sample(self, shape, dtype="float32", device=None):
        gaussian = np.random.normal(size=shape)
        gaussian = functional.convert_to_tensor(gaussian.astype("float32"), dtype=dtype)
        return gaussian.to(device) if backend.is_torch_backend else gaussian

    def __call__(
        self,
        prompt,
        image=None,  # [image_to_image and inpaint] None for `text_to_image` using prompt only, not None for `image_to_image` or `inpaint` mode
        batch_size=1,
        image_shape=(512, 512),  # int or list of 2 int, should be divisible by 64. Will be overwritten using `image.shape` if `image` not None
        repeat_noise=False,  # specified whether the noise should be same for all samples in the batch
        temperature=1,  # the noise temperature, random noise gets multiplied by this
        init_x0=None,  # tensor in same shape with `image_latents.shape`. Default `None` for random noise. Will be overwritten if `image` not None
        init_step=0,  # int value in `[0, self.num_steps==50]`, skip steps for denoising process. Will be overwritten using `strength` if `image` not None
        latent_scaling_factor=0.18215,  # scaling factor for the image latent space. Encoder outputs and Decoder inputs are scaled by this value
        uncond_scale=7.5,  # unconditional guidance scale: "eps = eps(x, empty) + scale * (eps(x, cond) - eps(x, empty))"
        return_inner=False,  # boolean value if return inner step results for visualizing the process
        strength=0.75,  # [image_to_image and inpaint] specifies how much of the original image should **NOT** be preserved
        inpaint_mask=None,  # [inpaint], in same shape with image_latents, or 4 float [top, left, bottom, right] in [0, 1], like [0.5, 0, 1, 1] for bottom half
    ):
        device = next(self.unet_model.parameters()).device if backend.is_torch_backend else None
        compute_dtype = self.unet_model.compute_dtype

        """ Checking if text_to_image / image_to_image / inpaint mode """
        is_inpaint = image is not None and inpaint_mask is not None
        if image is not None:  # image_to_image
            init_x0, init_step, image_shape, inpaint_orign_latents, inpaint_orign_noise = self.encode_image(image, batch_size, strength, latent_scaling_factor)
            print(">>>> Encoded image latents: {}, image_shape: {}, init_step: {}".format(init_x0.shape, image_shape, init_step))
        if is_inpaint:  # inpaint
            inpaint_mask = self.init_inpaint_mask(inpaint_mask, inpaint_orign_latents.shape)
            print(">>>> Inpaint, inpaint_mask.shape:", inpaint_mask.shape)

        """ Init x0 """
        batch_size = batch_size if init_x0 is None else init_x0.shape[0]
        image_shape = image_shape if isinstance(image_shape, (list, tuple)) else [image_shape, image_shape]
        if image_data_format() == "channels_last":
            target_shape = [batch_size, image_shape[0] // 8, image_shape[1] // 8, self.latents_input_shape[-1]]
        else:
            target_shape = [batch_size, self.latents_input_shape[1], image_shape[-2] // 8, image_shape[-1] // 8]

        if init_x0 is None:
            xt = self.gaussian_sample(target_shape, dtype="float32", device=device)
        else:
            xt = functional.convert_to_tensor(init_x0, dtype="float32")
            xt = xt.to(device) if backend.is_torch_backend else xt

        """ Init prompt latents using clip model """
        if self.uncond_prompt is None:
            uncond_prompt = functional.convert_to_tensor(self.caption_tokenizer("", padding_value=self.caption_tokenizer.eot_token)[None], dtype=compute_dtype)
            uncond_prompt = uncond_prompt.long().to(device) if backend.is_torch_backend else uncond_prompt
            self.uncond_token = self.clip_model(uncond_prompt)
        cond_prompt = functional.convert_to_tensor(self.caption_tokenizer(prompt, padding_value=self.caption_tokenizer.eot_token)[None], dtype=compute_dtype)
        cond_prompt = cond_prompt.long().to(device) if backend.is_torch_backend else cond_prompt
        cond_token = self.clip_model(cond_prompt)
        uncond_cond_prompt = functional.concat([self.uncond_token] * batch_size + [cond_token] * batch_size, axis=0)

        """ Denoising using UNet """
        gathered_inner = []  # return multiple if return_inner is True
        for cur_step in tqdm(range(self.num_steps - init_step)[::-1]):  # 50 -> 0
            time_step = functional.convert_to_tensor(np.stack([self.time_steps[cur_step]] * batch_size * 2), dtype="int64")
            time_step = time_step.to(device) if backend.is_torch_backend else time_step
            xt_inputs = functional.concat([xt, xt], axis=0)

            # get_eps
            out = self.unet_model([xt_inputs, time_step, uncond_cond_prompt])
            e_t_uncond, e_t_cond = functional.split(out, 2, axis=0)
            e_t = e_t_uncond + (e_t_cond - e_t_uncond) * uncond_scale

            # get_x_prev_and_pred_x0
            ddim_alpha_prev, ddim_sigma = self.ddim_alpha_prev[cur_step], self.ddim_sigma[cur_step]
            pred_x0 = (xt - e_t * self.ddim_sqrt_one_minus_alpha[cur_step]) / (self.ddim_alpha[cur_step] ** 0.5)  # Current prediction for x_0
            dir_xt = e_t * ((1.0 - ddim_alpha_prev - ddim_sigma**2) ** 0.5)  # Direction pointing to x_t

            noise_shape = (1, *target_shape[1:]) if repeat_noise else target_shape
            noise = 0.0 if ddim_sigma == 0 else self.gaussian_sample(noise_shape, dtype=pred_x0.dtype, device=device)
            xt = (ddim_alpha_prev**0.5) * pred_x0 + dir_xt + ddim_sigma * temperature * noise

            if is_inpaint:  # q_sample
                orig_t = self.ddim_alpha_sqrt[cur_step] * inpaint_orign_latents + self.ddim_sqrt_one_minus_alpha[cur_step] * inpaint_orign_noise
                xt = orig_t * inpaint_mask + xt * (1 - inpaint_mask)  # Replace the masked area with original image latents

            if return_inner:
                gathered_inner.append(xt)

        """ Decode the image """
        if return_inner:
            return [self.decoder_model(inner / latent_scaling_factor) for inner in tqdm(gathered_inner, "Decoding")]
        else:
            return self.decoder_model(xt / latent_scaling_factor)

    def encode_image(self, image, batch_size=1, strength=0.75, latent_scaling_factor=0.18215):
        """
        strength: specifies how much of the original image should not be preserved
        latent_scaling_factor: scaling factor for the image latent space. Encoder outputs and Decoder inputs are scaled by this value

        return init_x0, init_step, (height, width), image_latents, noise
        """
        device = next(self.unet_model.parameters()).device if backend.is_torch_backend else None
        compute_dtype = self.unet_model.compute_dtype

        if self.encoder_model is None:
            print(">>>> Build default Encoder model: stable_diffusion.Encoder")
            encoder_model = self.build_model("stable_diffusion.Encoder", **self.encoder_model_kwargs)
            self.encoder_model = encoder_model.to(device) if backend.is_torch_backend else encoder_model
            self.models.append(self.encoder_model)

        """ Load image """
        if isinstance(image, str):
            image = Image.open(image).convert("RGB")
            width, height = image.size[0] - image.size[0] % 64, image.size[1] - image.size[1] % 64  # or 32 [???]
            image = np.array(image.resize((width, height), resample=Image.Resampling.BICUBIC))
        else:  # np.array inputs
            height, width = image.shape[0] - image.shape[0] % 64, image.shape[1] - image.shape[1] % 64  # or 32 [???]
            image = image[:height, :width]  # Just crop
            image = image if image.max() > 2 else image * 255  # -> [0, 255]
        print(">>>> input image.shape: {}, image.min(): {}, image.max(): {}".format(image.shape, image.min(), image.max()))

        """ Encoder """
        image = image.astype("float32")[None] / 127.5 - 1
        image = image if image_data_format() == "channels_last" else image.transpose([0, 3, 1, 2])
        image = functional.convert_to_tensor(image, dtype=compute_dtype)
        image = image.to(device) if backend.is_torch_backend else image
        encoder_outputs = self.encoder_model(image)

        """ Gaussian sampling"""
        channel_axis = -1 if backend.image_data_format() == "channels_last" else 1
        mean, log_var = functional.split(encoder_outputs, 2, axis=channel_axis)
        log_var = functional.clip_by_value(log_var, -30.0, 20.0)
        std = functional.exp(log_var * 0.5)
        gaussian = self.gaussian_sample(std.shape, dtype=compute_dtype, device=device)
        image_latents = (mean + std * gaussian) * latent_scaling_factor
        image_latents = functional.concat([image_latents] * batch_size, axis=0)

        """ q_sample """
        noise = self.gaussian_sample(image_latents.shape, dtype=compute_dtype, device=device)
        timestep_start = int(strength * self.num_steps)
        init_x0 = self.ddim_alpha_sqrt[timestep_start] * image_latents + self.ddim_sqrt_one_minus_alpha[timestep_start] * noise
        init_step = self.num_steps - timestep_start
        return init_x0, init_step, (height, width), image_latents, noise

    def init_inpaint_mask(self, inpaint_mask, latents_shape):
        device = next(self.unet_model.parameters()).device if backend.is_torch_backend else None
        compute_dtype = self.unet_model.compute_dtype

        if isinstance(inpaint_mask, (list, tuple)) and len(inpaint_mask) == 4:
            height, width = latents_shape[1:-1] if image_data_format() == "channels_last" else latents_shape[2:]
            top, left, bottom, right = (np.array(inpaint_mask) * [height, width, height, width]).astype("int64")
            inpaint_mask = np.zeros(latents_shape).astype("float32")
            if image_data_format() == "channels_last":
                inpaint_mask[:, top:bottom, left:right, :] = 1
            else:
                inpaint_mask[:, :, top:bottom, left:right] = 1
            inpaint_mask = functional.convert_to_tensor(inpaint_mask.astype("float32"), dtype=compute_dtype)
            inpaint_mask = inpaint_mask.to(device) if backend.is_torch_backend else inpaint_mask
        else:
            assert inpaint_mask.shape == latents_shape, "provided inpaint_mask not in same shape with image_latents: {}".format(latents_shape)
            inpaint_mask = functional.convert_to_tensor(inpaint_mask.astype("float32"), dtype=compute_dtype)
            inpaint_mask = inpaint_mask.to(device) if backend.is_torch_backend else inpaint_mask
        return inpaint_mask



from keras_cv_attention_models.beit.beit import Beit, keras_model_load_weights_from_pytorch_model
from keras_cv_attention_models.models import register_model


def EVA(layer_scale=0, use_abs_pos_emb=True, model_name="eva", **kwargs):
    kwargs.pop("kwargs", None)
    patch_size = kwargs.pop("patch_size", 14)
    force_reload_mismatch = patch_size != 14  # If patch_size not 14, force reload pos_emb and stem_conv weights
    return Beit(**locals(), **kwargs)


@register_model
def EvaLargePatch14(input_shape=(196, 196, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet21k-ft1k", **kwargs):
    embed_dim = 1024
    depth = 24
    num_heads = 16
    attn_qkv_bias = True
    return EVA(**locals(), model_name="eva_large_patch14", **kwargs)


@register_model
def EvaGiantPatch14(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet21k-ft1k", **kwargs):
    mlp_ratio = 6144 / 1408
    embed_dim = 1408
    depth = 40
    num_heads = 16
    return EVA(**locals(), model_name="eva_giant_patch14", **kwargs)



from keras_cv_attention_models.beit.beit import (
    Beit,
    BeitBasePatch16,
    BeitLargePatch16,
    BeitV2BasePatch16,
    BeitV2LargePatch16,
    PatchConv2DWithResampleWeights,
    MultiHeadRelativePositionalEmbedding,
    PositionalEncodingFourierRot,
    keras_model_load_weights_from_pytorch_model,
)
from keras_cv_attention_models.beit.dinov2 import DINOv2, DINOv2_ViT_Small14, DINOv2_ViT_Base14, DINOv2_ViT_Large14, DINOv2_ViT_Giant14
from keras_cv_attention_models.beit.eva import EVA, EvaLargePatch14, EvaGiantPatch14
from keras_cv_attention_models.beit.eva02 import EVA02, EVA02TinyPatch14, EVA02SmallPatch14, EVA02BasePatch14, EVA02LargePatch14
from keras_cv_attention_models.beit.flexivit import FlexiViT, FlexiViTSmall, FlexiViTBase, FlexiViTLarge
from keras_cv_attention_models.beit.meta_transformer import MetaTransformer, MetaTransformerBasePatch16, MetaTransformerLargePatch14
from keras_cv_attention_models.beit.vit import ViT, ViTTinyPatch16, ViTBasePatch16, ViTLargePatch14, ViTText, ViTTextLargePatch14

__beit_head_doc__ = """
Keras implementation of [beit](https://github.com/microsoft/unilm/tree/master/beit).
Paper [PDF 2106.08254 BEIT: BERT Pre-Training of Image Transformers](https://arxiv.org/pdf/2106.08254.pdf).
"""

__beitv2_head_doc__ = """
Keras implementation of [Github microsoft/beit2](https://github.com/microsoft/unilm/tree/master/beit2).
Paper [PDF 2208.06366 BEiT v2: Masked Image Modeling with Vector-Quantized Visual Tokenizers](https://arxiv.org/pdf/2208.06366.pdf).
"""

__dinov2_head_doc__ = """
Keras implementation of [Github facebookresearch/dinov2](https://github.com/facebookresearch/dinov2).
Paper [PDF 2304.07193 DINOv2: Learning Robust Visual Features without Supervision](https://arxiv.org/pdf/2304.07193.pdf).
"""

__eva_head_doc__ = """
Keras implementation of [Github baaivision/EVA](https://github.com/baaivision/EVA).
Paper [PDF 2211.07636 EVA: Exploring the Limits of Masked Visual Representation Learning at Scale](https://arxiv.org/pdf/2211.07636.pdf).
"""

__eva02_head_doc__ = """
Keras implementation of [Github baaivision/EVA/EVA-02](https://github.com/baaivision/EVA/tree/master/EVA-02).
Paper [PDF 2303.11331 EVA: EVA-02: A Visual Representation for Neon Genesis](https://arxiv.org/pdf/2303.11331.pdf).
"""

__flexivit_head_doc__ = """
Keras implementation of [Github google-research/big_vision](https://github.com/google-research/big_vision/tree/main/big_vision/configs/proj/flexivit).
Paper [PDF 2212.08013 FlexiViT: One Model for All Patch Sizes](https://arxiv.org/pdf/2212.08013.pdf).
"""

__vit_head_doc__ = """
Keras implementation of [Github google-research/vision_transformer](https://github.com/google-research/vision_transformer).
Paper [PDF 2010.11929 An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale](https://arxiv.org/pdf/2010.11929.pdf).
"""

__meta_transformer_head_doc__ = """
Keras implementation of [Github invictus717/MetaTransformer](https://github.com/invictus717/MetaTransformer).
Paper [PDF 2307.10802 Meta-Transformer: A Unified Framework for Multimodal Learning](https://arxiv.org/abs/2307.10802).
"""

__tail_doc__ = """  patch_size: stem patch size. Default {patch_size}.
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  layer_scale: init value for `attention` and `mlp` branch `gamma`, if > 0 will use `layer_scale` on block output.
      Default 0 for Vit and EVA, 0.1 for Beit, 1.0 for DINOv2.
  activation: activation used in whole model, default `gelu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be a constant value like `0.2`,
      or a tuple value like `(0, 0.2)` indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `None`.
  pretrained: one of `None` (random initialization) or 'imagenet21k-ft1k' (pre-training on ImageNet21k and fine-tuned ImageNet).
      Will try to download and load pre-trained model weights if not None.
  force_reload_mismatch: boolean value, set True if `patch_size` changed, will force reloading `pos_emb` and `stem_conv` weights.

Returns:
    A `keras.Model` instance.
"""

__class_tail_doc__ = """
Args:
  depth: number of blocks. Default `12`.
  embed_dim: channel dimension for stem and all blocks. Default `768`.
  num_heads: heads number for transformer blocks. Default `12`.
  model_name: string, model name.

  [Attention args]
  attn_key_dim: key dimension for transformer blocks. Default `0`.
  attn_qv_bias: boolean value, if True and attn_qkv_bias being False, will add BiasLayer for `query` and `value` in transformer.
      Default False for Vit, True for Beit.
  attn_qkv_bias: boolean value, if True will just use bias in `qkv_dense`, and set `qv_bias` False.
      Default True for Vit, False for Beit.
  attn_out_weight: boolean value if use output dense for transformer. Default `True`.
  attn_out_bias: boolean value if output dense use bias for transformer. Default `True`.
  attn_dropout: `attention_score` dropout rate. Default `0`.
  use_abs_pos_emb: boolean value if use abcolute positional embedding or relative one in attention blocks.
      Default True for Vit, False for Beit.
  use_abs_pos_emb_on_cls_token: boolean value, if `use_abs_pos_emb` is True, whether apply `pos_emb` on `cls_token`.
      False for `FlexiViT`, same as `no_embed_class` in timm. Default True for others.
  use_rot_pos_emb: boolean value if use `PositionalEncodingFourierRot` on attention query and key.
      True for EVA02, False for others.

  [MLP args]
  mlp_ratio: dimension expansion ration for `mlp_block`s. Default `4`.
  use_gated_mlp: boolean value if using `dense gated + swish` instead of `activation` in `mlp` block.
      True for DINOv2 and EVA02, False for others.
  use_norm_mlp: boolean value if use additional LayerNorm for MLP block. True for EVA02 base and large, False for others.

  [Head args]
  use_mean_pooling_head: boolean value if use mean output or `class_token` output. Default False for Vit, True for Beit.
  use_cat_head: boolean value if use mean output concatenated with `class_token` output. Default True for DINOv2, False for others.

  [Text model args]
  vocab_size: set value > 0 for building text model. Default 49408 for ViTText, 0 for others.
  max_block_size: max block size, works only if vocab_size > 0. Default 77.
  text_positional_dropout: dropout for text model embedding layers. Default 0.
  text_use_positional_embedding: boolean value if use Embedding positional layer after inputs. Default True.
  include_top: boolean value if include top output Dense layer, True for using output channles == vocab_size. Default True.

  [common args]
""" + __tail_doc__

Beit.__doc__ = __beit_head_doc__ + __class_tail_doc__.format(patch_size=16) + """
Model architectures:
  | Model                 | Params  | FLOPs   | Input | Top1 Acc |
  | --------------------- | ------- | ------- | ----- | -------- |
  | BeitBasePatch16, 21k  | 86.53M  | 17.61G  | 224   | 85.240   |
  |                       | 86.74M  | 55.70G  | 384   | 86.808   |
  | BeitLargePatch16, 21k | 304.43M | 61.68G  | 224   | 87.476   |
  |                       | 305.00M | 191.65G | 384   | 88.382   |
  |                       | 305.67M | 363.46G | 512   | 88.584   |

  | Model              | Params  | FLOPs  | Input | Top1 Acc |
  | ------------------ | ------- | ------ | ----- | -------- |
  | BeitV2BasePatch16  | 86.53M  | 17.61G | 224   | 85.5     |
  | - imagenet21k-ft1k | 86.53M  | 17.61G | 224   | 86.5     |
  | BeitV2LargePatch16 | 304.43M | 61.68G | 224   | 87.3     |
  | - imagenet21k-ft1k | 304.43M | 61.68G | 224   | 88.4     |
"""

DINOv2.__doc__ = __dinov2_head_doc__ + __class_tail_doc__.format(patch_size=14) + """
Model architectures:
  | Model              | Params  | FLOPs   | Input | Top1 Acc |
  | ------------------ | ------- | ------- | ----- | -------- |
  | DINOv2_ViT_Small14 | 22.83M  | 47.23G  | 518   | 81.1     |
  | DINOv2_ViT_Base14  | 88.12M  | 152.6G  | 518   | 84.5     |
  | DINOv2_ViT_Large14 | 306.4M  | 509.6G  | 518   | 86.3     |
  | DINOv2_ViT_Giant14 | 1139.6M | 1790.3G | 518   | 86.5     |
"""

EVA.__doc__ = __eva_head_doc__ + __class_tail_doc__.format(patch_size=14) + """
Model architectures:
  | Model                 | Params  | FLOPs    | Input | Top1 Acc |
  | --------------------- | ------- | -------- | ----- | -------- |
  | EvaLargePatch14, 22k  | 304.14M | 61.65G   | 196   | 88.59    |
  |                       | 304.53M | 191.55G  | 336   | 89.20    |
  | EvaGiantPatch14, clip | 1012.6M | 267.40G  | 224   | 89.10    |
  | - m30m                | 1013.0M | 621.45G  | 336   | 89.57    |
  | - m30m                | 1014.4M | 1911.61G | 560   | 89.80    |
"""

EVA02.__doc__ = __eva02_head_doc__ + __class_tail_doc__.format(patch_size=14) + """
Model architectures:
  | Model                                  | Params  | FLOPs   | Input | Top1 Acc |
  | -------------------------------------- | ------- | ------- | ----- | -------- |
  | EVA02TinyPatch14, mim_in22k_ft1k       | 5.76M   | 4.72G   | 336   | 80.658   |
  | EVA02SmallPatch14, mim_in22k_ft1k      | 22.13M  | 15.57G  | 336   | 85.74    |
  | EVA02BasePatch14, mim_in22k_ft22k_ft1k | 87.12M  | 107.6G  | 448   | 88.692   |
  | EVA02LargePatch14, mim_m38m_ft22k_ft1k | 305.08M | 363.68G | 448   | 90.054   |
"""

FlexiViT.__doc__ = __flexivit_head_doc__ + __class_tail_doc__.format(patch_size=16) + """
Model architectures:
  | Model         | Params  | FLOPs  | Input | Top1 Acc |
  | ------------- | ------- | ------ | ----- | -------- |
  | FlexiViTSmall | 22.06M  | 5.36G  | 240   | 82.53    |
  | FlexiViTBase  | 86.59M  | 20.33G | 240   | 84.66    |
  | FlexiViTLarge | 304.47M | 71.09G | 240   | 85.64    |
"""

ViT.__doc__ = __vit_head_doc__ + __class_tail_doc__.format(patch_size=16)
ViTText.__doc__ = __vit_head_doc__ + __class_tail_doc__.format(patch_size=16)

MetaTransformer.__doc__ = __meta_transformer_head_doc__ + __class_tail_doc__.format(patch_size=16) + """
Model architectures:
  | Model                                 | Params  | FLOPs  | Input | Top1 Acc |
  | ------------------------------------- | ------- | ------ | ----- | -------- |
  | MetaTransformerBasePatch16, laion_2b  | 86.86M  | 55.73G | 384   | 85.4     |
  | MetaTransformerLargePatch14, laion_2b | 304.53M | 191.6G | 336   | 88.1     |
"""

__model_tail_doc__ = """
Args:
""" + __tail_doc__

BeitBasePatch16.__doc__ = __beit_head_doc__ + __model_tail_doc__.format(patch_size=16)
BeitLargePatch16.__doc__ = BeitBasePatch16.__doc__

BeitV2BasePatch16.__doc__ = __beitv2_head_doc__ + __model_tail_doc__.format(patch_size=16)
BeitV2LargePatch16.__doc__ = BeitV2BasePatch16.__doc__

DINOv2_ViT_Small14.__doc__ = __dinov2_head_doc__ + __model_tail_doc__.format(patch_size=14)
DINOv2_ViT_Base14.__doc__ = DINOv2_ViT_Small14.__doc__
DINOv2_ViT_Large14.__doc__ = DINOv2_ViT_Small14.__doc__
DINOv2_ViT_Giant14.__doc__ = DINOv2_ViT_Small14.__doc__

EvaLargePatch14.__doc__ = __eva_head_doc__ + __model_tail_doc__.format(patch_size=14)
EvaGiantPatch14.__doc__ = EvaLargePatch14.__doc__

EVA02TinyPatch14.__doc__ = __eva02_head_doc__ + __model_tail_doc__.format(patch_size=14)
EVA02SmallPatch14.__doc__ = EVA02TinyPatch14.__doc__
EVA02BasePatch14.__doc__ = EVA02TinyPatch14.__doc__
EVA02LargePatch14.__doc__ = EVA02TinyPatch14.__doc__

FlexiViTSmall.__doc__ = __flexivit_head_doc__ + __model_tail_doc__.format(patch_size=16)
FlexiViTBase.__doc__ = FlexiViTSmall.__doc__
FlexiViTLarge.__doc__ = FlexiViTSmall.__doc__

MetaTransformerBasePatch16.__doc__ = __meta_transformer_head_doc__ + __model_tail_doc__.format(patch_size=16)
MetaTransformerLargePatch14.__doc__ = __meta_transformer_head_doc__ + __model_tail_doc__.format(patch_size=14)

ViTTinyPatch16.__doc__ = __vit_head_doc__ + __model_tail_doc__.format(patch_size=16)
ViTBasePatch16.__doc__ = __vit_head_doc__ + __model_tail_doc__.format(patch_size=16)
ViTLargePatch14.__doc__ = __vit_head_doc__ + __model_tail_doc__.format(patch_size=14)

MultiHeadRelativePositionalEmbedding.__doc__ = __beit_head_doc__ + """
Multi Head Relative Positional Embedding layer.

Positional embedding shape is `[num_heads, (2 * height - 1) * (2 * width - 1)]`.
input (with_cls_token=True): `[batch, num_heads, attn_blocks, attn_blocks]`. where `attn_blocks = attn_height * attn_width + class_token`
input (with_cls_token=False): `[batch, num_heads, attn_blocks, attn_blocks]`. where `attn_blocks = attn_height * attn_width`
output: `[batch, num_heads, attn_blocks, attn_blocks] + positional_bias`.
conditions: attn_height == attn_width

Args:
  with_cls_token: boolean value if input is with class_token.
  attn_height: specify `height` for `attn_blocks` if not square `attn_height != attn_width`.
  num_heads: specify num_heads, or using `input.shape[1]`.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> aa = attention_layers.MultiHeadRelativePositionalEmbedding()
>>> print(f"{aa(tf.ones([1, 8, 29 * 29 + 1, 29 * 29 + 1])).shape = }")
# aa(tf.ones([1, 8, 29 * 29 + 1, 29 * 29 + 1])).shape = TensorShape([1, 8, 842, 842])
>>> print({ii.name:ii.numpy().shape for ii in aa.weights})
# {'multi_head_relative_positional_embedding/pos_emb:0': (3252, 8)}

>>> aa = attention_layers.MultiHeadRelativePositionalEmbedding(with_cls_token=False)
>>> print(f"{aa(tf.ones([1, 8, 29 * 29, 29 * 29])).shape = }")
# aa(tf.ones([1, 8, 29 * 29, 29 * 29])).shape = TensorShape([1, 8, 841, 841])
>>> print({ii.name:ii.numpy().shape for ii in aa.weights})
# {'multi_head_relative_positional_embedding_1/pos_emb:0': (3249, 8)}

>>> plt.imshow(aa.relative_position_index)
"""

PositionalEncodingFourierRot.__doc__ = __eva02_head_doc__ + """
Positional Encoding Fourier Rot. No weight.

Positional embedding shape is `pos_sin = pos_cos = [attn_height * attn_width, channels]`.
input (with_cls_token=True): `[batch, ...,  attn_height * attn_width + class_token, channels]`.
input (with_cls_token=False): `[batch, ..., attn_height * attn_width, channels]`.
output: `input * pos_cos + rot * pos_sin`.

Args:
  with_cls_token: boolean value if input is with class_token.
  attn_height: specify `height` for `attn_blocks` if not square `attn_height != attn_width`.
  num_heads: specify num_heads.
      - if <=0, will calculate pos_sin and pos_cos depends on `inputs[-1]`. Default behavior.
      - if > 0, will calculate pos_sin and pos_cos depends on `inputs[-1] // num_heads`,
        then repeat and reshape to match `inputs[-1]`.
  temperature: temperature.
  ref_feature_shape: reference feature shape for resize / fine-tune

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> aa = attention_layers.PositionalEncodingFourierRot()
>>> print(f"{aa(tf.ones([1, 29 * 29 + 1, 3 * 64])).shape = }")
# aa(tf.ones([1, 29 * 29 + 1, 3 * 64])).shape = TensorShape([1, 842, 192])
>>> print(f"{aa.pos_sin.shape = }, {aa.pos_cos.shape = }")
# aa.pos_sin.shape = TensorShape([841, 192]), aa.pos_cos.shape = TensorShape([841, 192])

# with_cls_token=False
>>> aa = attention_layers.PositionalEncodingFourierRot(with_cls_token=False)
>>> print(f"{aa(tf.ones([1, 8, 29 * 29, 3 * 64])).shape = }")
# aa(tf.ones([1, 8, 29 * 29, 3 * 64])).shape = TensorShape([1, 8, 841, 192])
>>> print(f"{aa.pos_sin.shape = }, {aa.pos_cos.shape = }")
# aa.pos_sin.shape = TensorShape([841, 192]), aa.pos_cos.shape = TensorShape([841, 192])

# with_cls_token=False, num_heads=3, pos_sin and pos_cos is repeated 3 times.
>>> aa = attention_layers.PositionalEncodingFourierRot(with_cls_token=False, num_heads=3)
>>> print(f"{aa(tf.ones([1, 8, 29 * 29, 3 * 64])).shape = }")
# aa(tf.ones([1, 8, 29 * 29, 3 * 64])).shape = TensorShape([1, 8, 841, 192])
>>> print(f"{aa.pos_sin.shape = }, {aa.pos_cos.shape = }")
# aa.pos_sin.shape = TensorShape([841, 192]), aa.pos_cos.shape = TensorShape([841, 192])
>>> print(f"{np.allclose(aa.pos_sin[:, :64], aa.pos_sin[:, -64:]) = }")
# np.allclose(aa.pos_sin[:, :64], aa.pos_sin[:, -64:]) = True
"""

PatchConv2DWithResampleWeights.__doc__ = __flexivit_head_doc__ + """
Source implementation https://github.com/google-research/big_vision/blob/main/big_vision/models/proj/flexi/vit.py#L30
Typical Conv2D layer with `load_resized_weights` function, also using PyTorch padding style.

Args: Same with Conv2D.
"""

keras_model_load_weights_from_pytorch_model.__doc__ = """
Keras ViT model reload weights from timm torch model using `keras_cv_attention_models.download_and_load.keras_reload_from_torch_model`.
Can be used for loading most `Beit` / `ViT` / `FlexViT` / `EVA` models.
Parameters in this using of `keras_reload_from_torch_model` is specifically fitted for timm ViT models.

Args:
  keras_model: keras model same architecture with `timm_vit_model`.
      Need to set parameters for `keras_cv_attention_models.vit.ViT` same as torch model.
  timm_vit_model: built timm vit model.
  save_name: output file name. If ends with `.h5` will save a h5 model, or will be saved_model format.
      Default None for using `{model_name}_{input_shape}.h5`.

Outputs:
  Converted `save_name` weight file, which can be used for afterward and further usage.

Examples:
>>> # Build a timm ViT model
>>> import timm
>>> torch_model = timm.models.vit_tiny_patch16_224(pretrained=True)
>>> _ = torch_model.eval()
>>>
>>> # Build a ViT model same architecture with torch_model
>>> from keras_cv_attention_models import vit
>>> mm = vit.ViT(depth=12, embed_dim=192, num_heads=3, pretrained=None, classifier_activation=None)
>>> vit.keras_model_load_weights_from_pytorch_model(mm, torch_model)
# >>>> Save model to: vit_224.h5
# >>>> Keras model prediction: [('n02123045', 'tabby', 11.990417), ('n02123159', 'tiger_cat', 11.630723), ...]
# >>>> Torch model prediction: [[('n02123045', 'tabby', 11.99042), ('n02123159', 'tiger_cat', 11.630725), ...]
"""



from keras_cv_attention_models.nat.nat import NAT
from keras_cv_attention_models.models import register_model


@register_model
def DiNAT_Mini(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 6, 5]
    # dilation_rates = [[1, 8, 1], [1, 4, 1, 4], [1, 2] * 3, [1, 1, 1, 1, 1]]
    use_every_other_dilations = True
    return NAT(**locals(), model_name="dinat_mini", **kwargs)


@register_model
def DiNAT_Tiny(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 18, 5]
    # dilation_rates = [[1, 8, 1], [1, 4, 1, 4], [1, 2] * 9, [1, 1, 1, 1, 1]]
    use_every_other_dilations = True
    return NAT(**locals(), model_name="dinat_tiny", **kwargs)


@register_model
def DiNAT_Small(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 18, 5]
    num_heads = [3, 6, 12, 24]
    out_channels = [96, 192, 384, 768]
    mlp_ratio = kwargs.pop("mlp_ratio", 2)
    # layer_scale = kwargs.pop("layer_scale", 1e-5)
    # dilation_rates = [[1, 8, 1], [1, 4, 1, 4], [1, 2] * 9, [1, 1, 1, 1, 1]]
    use_every_other_dilations = True
    return NAT(**locals(), model_name="dinat_small", **kwargs)


@register_model
def DiNAT_Base(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 18, 5]
    num_heads = [4, 8, 16, 32]
    out_channels = [128, 256, 512, 1024]
    mlp_ratio = kwargs.pop("mlp_ratio", 2)
    layer_scale = kwargs.pop("layer_scale", 1e-5)
    # dilation_rates = [[1, 8, 1], [1, 4, 1, 4], [1, 2] * 9, [1, 1, 1, 1, 1]]
    use_every_other_dilations = True
    return NAT(**locals(), model_name="dinat_base", **kwargs)


@register_model
def DiNAT_Large(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet21k-ft1k", **kwargs):
    num_blocks = [3, 4, 18, 5]
    num_heads = [6, 12, 24, 48]
    out_channels = [192, 384, 768, 1536]
    mlp_ratio = kwargs.pop("mlp_ratio", 2)
    # layer_scale = kwargs.pop("layer_scale", 1e-5)
    # dilation_rates = [[1, 8, 1], [1, 4, 1, 4], [1, 2] * 9, [1, 1, 1, 1, 1]]
    use_every_other_dilations = True
    return NAT(**locals(), model_name="dinat_large", **kwargs)


@register_model
def DiNAT_Large_K11(input_shape=(384, 384, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet21k-ft1k", **kwargs):
    num_blocks = [3, 4, 18, 5]
    num_heads = [6, 12, 24, 48]
    out_channels = [192, 384, 768, 1536]
    mlp_ratio = kwargs.pop("mlp_ratio", 2)
    # layer_scale = kwargs.pop("layer_scale", 1e-5)
    # dilation_rates = [[1, 8, 1], [1, 4, 1, 4], [1, 2] * 9, [1, 1, 1, 1, 1]]
    use_every_other_dilations = True
    attn_kernel_size = 11
    return NAT(**locals(), model_name="dinat_large_k11", **kwargs)



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, models, is_channels_last
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    add_with_layer_scale_and_drop_block,
    batchnorm_with_activation,
    conv2d_no_bias,
    ChannelAffine,
    depthwise_conv2d_no_bias,
    drop_block,
    MultiHeadPositionalEmbedding,
    qkv_to_multi_head_channels_last_format,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "efficientformer_v2_l": {"imagenet": {224: "3792a3ea9eb9d1e818d4a36c00e422a5"}},
    "efficientformer_v2_s0": {"imagenet": {224: "88de5ba4a8effd887d53df3020ba8433"}},
    "efficientformer_v2_s1": {"imagenet": {224: "a0843565d0d01004604d52b0e2ddfa0a"}},
    "efficientformer_v2_s2": {"imagenet": {224: "07c358cc0f8ea8a02722673bd38bfe97"}},
}


def conv_mhsa_with_multi_head_position(
    inputs,
    num_heads=8,
    key_dim=32,
    strides=1,
    out_shape=None,
    attn_ratio=4,
    use_local_global_query=False,
    use_talking_head=True,
    qkv_bias=True,
    activation="gelu",
    name=None,
):
    height_axis, width_axis, channel_aixs = (1, 2, 3) if is_channels_last() else (2, 3, 1)
    height, width, input_channel = inputs.shape[height_axis], inputs.shape[width_axis], inputs.shape[channel_aixs]
    key_dim = key_dim if key_dim > 0 else input_channel // num_heads
    # qk_scale = float(1.0 / tf.math.sqrt(tf.cast(key_dim, "float32")))
    qk_scale = 1.0 / (float(key_dim) ** 0.5)
    out_shape = input_channel if out_shape is None else out_shape
    qk_out = num_heads * key_dim
    value_out = attn_ratio * qk_out
    value_dim = attn_ratio * key_dim

    if strides > 1:
        should_cut_height, should_cut_width = height % 2, width % 2  # keep shape same with inputs after later UpSampling2D
        inputs = depthwise_conv2d_no_bias(inputs, use_bias=True, kernel_size=3, strides=strides, padding="same", name=name and name + "down_sample_")
        inputs = batchnorm_with_activation(inputs, activation=None, name=name and name + "down_sample_")

    kv_blocks = height * width

    if use_local_global_query:
        # pool_query = layers.AvgPool2D(pool_size=1, strides=2)(inputs)
        # pool_query = inputs[:, ::2, ::2] if is_channels_last() else inputs[:, :, ::2, ::2] # nn.AvgPool2d(kernel_size=1, stride=2, padding=0)
        pool_query = layers.AvgPool2D(pool_size=1, strides=2)(inputs)
        local_query = depthwise_conv2d_no_bias(inputs, use_bias=qkv_bias, kernel_size=3, strides=2, padding="same", name=name and name + "local_query_")
        pre_query = pool_query + local_query
        vv_local_strides = 2
    else:
        pre_query = inputs
        vv_local_strides = 1
    query_height, query_width = pre_query.shape[1:-1] if is_channels_last() else pre_query.shape[2:]

    query = conv2d_no_bias(pre_query, qk_out, use_bias=qkv_bias, kernel_size=1, name=name and name + "query_")
    query = batchnorm_with_activation(query, activation=None, name=name and name + "query_")

    key = conv2d_no_bias(inputs, qk_out, use_bias=qkv_bias, kernel_size=1, name=name and name + "key_")
    key = batchnorm_with_activation(key, activation=None, name=name and name + "key_")

    value = conv2d_no_bias(inputs, value_out, use_bias=qkv_bias, kernel_size=1, name=name and name + "value_")
    value = batchnorm_with_activation(value, activation=None, name=name and name + "value_")
    vv_local = depthwise_conv2d_no_bias(value, use_bias=qkv_bias, kernel_size=3, strides=vv_local_strides, padding="same", name=name and name + "value_local_")
    vv_local = batchnorm_with_activation(vv_local, activation=None, name=name and name + "value_local_")

    # query = functional.transpose(
    #     functional.reshape(query, [-1, query_height * query_width, num_heads, key_dim]), [0, 2, 1, 3]
    # )  #  [batch, num_heads, hh * ww, key_dim]
    # key = functional.transpose(functional.reshape(key, [-1, kv_blocks, num_heads, key_dim]), [0, 2, 3, 1])  #  [batch, num_heads, key_dim, hh * ww]
    # value = functional.transpose(functional.reshape(value, [-1, kv_blocks, num_heads, value_dim]), [0, 2, 1, 3])  #  [batch, num_heads, hh * ww, value_dim]
    query, key, value = qkv_to_multi_head_channels_last_format(query, key, value, num_heads=num_heads)

    # attention_scores = layers.Lambda(lambda xx: functional.matmul(xx[0], xx[1]))([query, key]) * qk_scale  # [batch, num_heads, hh * ww, hh * ww]
    # print(f"{query.shape = }, {key.shape = }, {value.shape = }, {attention_scores.shape = }, {query_height = }")
    attention_scores = (query @ key) * qk_scale
    attention_scores = MultiHeadPositionalEmbedding(query_height=query_height, name=name and name + "pos_emb")(attention_scores)

    if use_talking_head:
        # if channels_last, attention_scores [batch, num_heads, hh * ww, hh * ww] -> [batch, hh * ww, hh * ww, num_heads]
        attention_scores = functional.transpose(attention_scores, [0, 2, 3, 1]) if is_channels_last() else attention_scores
        attention_scores = conv2d_no_bias(attention_scores, num_heads, use_bias=True, name=name and name + "talking_head_1_")
        # On previous last dimension
        attention_scores = layers.Softmax(axis=2 if is_channels_last() else -1, name=name and name + "attention_scores")(attention_scores)
        attention_scores = conv2d_no_bias(attention_scores, num_heads, use_bias=True, name=name and name + "talking_head_2_")
        # if channels_last, attention_scores [batch, hh * ww, hh * ww, num_heads] -> [batch, num_heads, hh * ww, hh * ww]
        attention_scores = functional.transpose(attention_scores, [0, 3, 1, 2]) if is_channels_last() else attention_scores
    else:
        attention_scores = layers.Softmax(axis=-1, name=name and name + "attention_scores")(attention_scores)
        # attention_scores = layers.Dropout(attn_dropout, name=name and name + "attn_drop")(attention_scores) if attn_dropout > 0 else attention_scores

    # value = [batch, num_heads, hh * ww, value_dim], attention_output = [batch, num_heads, hh * ww, value_dim]
    # attention_output = layers.Lambda(lambda xx: functional.matmul(xx[0], xx[1]))([attention_scores, value])
    attention_output = attention_scores @ value
    if is_channels_last():
        attention_output = functional.transpose(attention_output, [0, 2, 1, 3])
        attention_output = functional.reshape(attention_output, [-1, query_height, query_width, num_heads * value_dim])
    else:
        attention_output = functional.transpose(attention_output, [0, 1, 3, 2])
        attention_output = functional.reshape(attention_output, [-1, num_heads * value_dim, query_height, query_width])
    attention_output += vv_local
    # print(f">>>> {attention_output.shape = }, {attention_scores.shape = }")

    if strides > 1:
        attention_output = layers.UpSampling2D(size=(strides, strides), interpolation="bilinear")(attention_output)
        if should_cut_height > 0 or should_cut_width > 0:
            keep_height, keep_width = attention_output.shape[height_axis] - should_cut_height, attention_output.shape[width_axis] - should_cut_width
            attention_output = attention_output[:, :keep_height, :keep_width] if is_channels_last() else attention_output[:, :, :keep_height, :keep_width]

    # [batch, hh, ww, num_heads * value_dim] * [num_heads * value_dim, out] --> [batch, hh, ww, out]
    attention_output = activation_by_name(attention_output, activation=activation, name=name)
    attention_output = conv2d_no_bias(attention_output, out_shape, use_bias=True, kernel_size=1, name=name and name + "output_")
    attention_output = batchnorm_with_activation(attention_output, activation=None, name=name and name + "output_")

    return attention_output


def mlp_block_with_additional_depthwise_conv(inputs, hidden_dim, output_channel=-1, drop_rate=0, activation="gelu", name=""):
    output_channel = output_channel if output_channel > 0 else inputs.shape[-1 if is_channels_last() else 1]
    nn = conv2d_no_bias(inputs, hidden_dim, kernel_size=1, use_bias=True, name=name + "1_")
    nn = batchnorm_with_activation(nn, activation=activation, name=name + "1_")

    nn = depthwise_conv2d_no_bias(nn, use_bias=True, kernel_size=3, strides=1, padding="same", name=name + "mid_")
    nn = batchnorm_with_activation(nn, activation=activation, name=name + "mid_")

    nn = conv2d_no_bias(nn, output_channel, kernel_size=1, use_bias=True, name=name + "2_")
    nn = batchnorm_with_activation(nn, activation=None, name=name + "2_")
    nn = layers.Dropout(drop_rate)(nn) if drop_rate > 0 else nn
    return nn


def down_sample_block(inputs, out_channel, use_attn=False, activation="gelu", name=""):
    conv_branch = conv2d_no_bias(inputs, out_channel, kernel_size=3, strides=2, use_bias=True, padding="same", name=name)
    conv_branch = batchnorm_with_activation(conv_branch, activation=None, name=name)
    if use_attn:
        fixed_kwargs = {"num_heads": 8, "key_dim": 16, "strides": 1, "attn_ratio": 4, "use_local_global_query": True, "use_talking_head": False}
        attn_branch = conv_mhsa_with_multi_head_position(inputs, **fixed_kwargs, out_shape=out_channel, activation=activation, name=name + "attn_")
        nn = attn_branch + conv_branch
    else:
        nn = conv_branch
    return nn


def EfficientFormerV2(
    num_blocks=[2, 2, 6, 4],
    out_channels=[32, 48, 96, 176],
    mlp_ratios=[4, 4, [4, 3, 3, 3, 4, 4], [4, 3, 3, 4]],
    num_attn_blocks_each_stack=[0, 0, 2, 2],
    stem_width=-1,
    stem_activation=None,
    layer_scale=1e-5,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    classifier_activation=None,
    use_distillation=False,
    dropout=0,
    pretrained=None,
    model_name="efficientformer_v2",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    stem_width = stem_width if stem_width > 0 else out_channels[0]
    stem_activation = stem_activation if stem_activation is not None else activation
    nn = conv2d_no_bias(inputs, stem_width // 2, 3, strides=2, use_bias=True, padding="same", name="stem_1_")
    nn = batchnorm_with_activation(nn, activation=stem_activation, name="stem_1_")
    nn = conv2d_no_bias(nn, stem_width, 3, strides=2, use_bias=True, padding="same", name="stem_2_")
    nn = batchnorm_with_activation(nn, activation=stem_activation, name="stem_2_")

    """ stages """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel) in enumerate(zip(num_blocks, out_channels)):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            use_attn = True if stack_id >= 3 else False  # Only the last stack
            nn = down_sample_block(nn, out_channel, use_attn=use_attn, activation=activation, name=stack_name + "downsample_")

        cur_num_attn_blocks = num_attn_blocks_each_stack[stack_id] if isinstance(num_attn_blocks_each_stack, (list, tuple)) else num_attn_blocks_each_stack
        cur_mlp_ratios = mlp_ratios[stack_id] if isinstance(mlp_ratios, (list, tuple)) else mlp_ratios
        for block_id in range(num_block):
            block_name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            cur_mlp_ratio = cur_mlp_ratios[block_id] if isinstance(cur_mlp_ratios, (list, tuple)) else cur_mlp_ratios
            if block_id > num_block - cur_num_attn_blocks - 1:
                strides = 2 if stack_id == 2 else 1
                attn_out = conv_mhsa_with_multi_head_position(nn, num_heads=8, key_dim=32, strides=strides, activation=activation, name=block_name + "attn_")
                nn = add_with_layer_scale_and_drop_block(nn, attn_out, layer_scale=layer_scale, drop_rate=block_drop_rate, name=block_name + "attn_")
            mlp_out = mlp_block_with_additional_depthwise_conv(nn, out_channel * cur_mlp_ratio, activation=activation, name=block_name + "mlp_")
            nn = add_with_layer_scale_and_drop_block(nn, mlp_out, layer_scale=layer_scale, drop_rate=block_drop_rate, name=block_name + "mlp_")
            global_block_id += 1

    """ output """
    if num_classes > 0:
        nn = batchnorm_with_activation(nn, activation=None, name="pre_output_")
        nn = layers.GlobalAveragePooling2D()(nn)  # tf.reduce_mean(nn, axis=1)
        if dropout > 0 and dropout < 1:
            nn = layers.Dropout(dropout)(nn)
        out = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="head")(nn)

        if use_distillation:
            distill = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="distill_head")(nn)
            out = [out, distill]
    else:
        out = nn

    model = models.Model(inputs, out, name=model_name)
    reload_model_weights(model, PRETRAINED_DICT, "efficientformer", pretrained, MultiHeadPositionalEmbedding)

    add_pre_post_process(model, rescale_mode="torch")
    model.switch_to_deploy = lambda: switch_to_deploy(model)
    return model


def switch_to_deploy(model):
    from keras_cv_attention_models.model_surgery.model_surgery import fuse_distill_head

    new_model = fuse_distill_head(model, head_bn="pre_output_bn", distill_head_bn="pre_output_bn") if "head" in model.output_names else model
    add_pre_post_process(new_model, rescale_mode=model.preprocess_input.rescale_mode, post_process=model.decode_predictions)
    return new_model


@register_model
def EfficientFormerV2S0(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", use_distillation=False, pretrained="imagenet", **kwargs):
    return EfficientFormerV2(**locals(), model_name="efficientformer_v2_s0", **kwargs)


@register_model
def EfficientFormerV2S1(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", use_distillation=False, pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 9, 6]
    out_channels = [32, 48, 120, 224]
    mlp_ratios = [4, 4, [4, 4, 3, 3, 3, 3, 4, 4, 4], [4, 4, 3, 3, 4, 4]]
    return EfficientFormerV2(**locals(), model_name="efficientformer_v2_s1", **kwargs)


@register_model
def EfficientFormerV2S2(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", use_distillation=False, pretrained="imagenet", **kwargs):
    num_blocks = [4, 4, 12, 8]
    out_channels = [32, 64, 144, 288]
    mlp_ratios = [4, 4, [4, 4, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4], [4, 4, 3, 3, 3, 3, 4, 4]]
    num_attn_blocks_each_stack = [0, 0, 4, 4]
    return EfficientFormerV2(**locals(), model_name="efficientformer_v2_s2", **kwargs)


@register_model
def EfficientFormerV2L(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", use_distillation=False, pretrained="imagenet", **kwargs):
    num_blocks = [5, 5, 15, 10]
    out_channels = [40, 80, 192, 384]
    mlp_ratios = [4, 4, [4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4], [4, 4, 4, 3, 3, 3, 3, 4, 4, 4]]
    num_attn_blocks_each_stack = [0, 0, 6, 6]
    return EfficientFormerV2(**locals(), model_name="efficientformer_v2_l", **kwargs)



from keras_cv_attention_models import backend as __backend__
from keras_cv_attention_models.stable_diffusion.stable_diffusion import StableDiffusion
from keras_cv_attention_models.stable_diffusion.unet import UNet, UNetTest
from keras_cv_attention_models.stable_diffusion.encoder_decoder import Encoder, Decoder
from keras_cv_attention_models.stable_diffusion.eval_func import RunPrediction

if __backend__.is_tensorflow_backend:
    from keras_cv_attention_models.stable_diffusion.data import build_tf_dataset as build_dataset
else:
    from keras_cv_attention_models.stable_diffusion.data import build_torch_dataset as build_dataset


__head_doc__ = """
Keras implementation of [Github CompVis/stable-diffusion](https://github.com/CompVis/stable-diffusion).
Paper [PDF 2112.10752 High-Resolution Image Synthesis with Latent Diffusion Models](https://arxiv.org/pdf/2112.10752.pdf).
"""

__tail_doc__ = """  image_shape: model image input shape and generated image shape.
      Should have exactly 3 inputs channels like `(224, 224, 3)`.
      Inner latents inpuit shape for UNet and Decode is `[image_shape[0] // 8, image_shape[1] // 8, 4]`.
  clip_model: str value like `beit.ViTTextLargePatch14` for models from this package under `keras_cv_attention_models`.
      Also can be a built model, or None for not using.
  unet_model: str value like `stable_diffusion.UNet` for models from this package under `keras_cv_attention_models`.
      Also can be a built model, or None for not using.
  decoder_model: str value like `stable_diffusion.Decoder` for models from this package under `keras_cv_attention_models`.
      Also can be a built model, or None for not using.
  encoder_model: str value like `stable_diffusion.Encoder` for models from this package under `keras_cv_attention_models`.
      Also can be a built model, or None for not using.
  clip_model_kwargs: dict value for kwargs used for building `clip_model`.
  unet_model_kwargs: dict value for kwargs used for building `unet_model`.
  decoder_model_kwargs: dict value for kwargs used for building `decoder_model`.
  encoder_model_kwargs: dict value for kwargs used for building `encoder_model`.
  caption_tokenizer: str value in ['GPT2Tokenizer', 'SimpleTokenizer', 'SentencePieceTokenizer'],
      or tiktoken one ['gpt2', 'r50k_base', 'p50k_base', 'cl100k_base'],
      or specified built tokenizer.
  num_steps: int value for the number of DDIM sampling steps, also means total denoising steps.
  num_training_steps: int value for total denoising steps during training.
  ddim_discretize: one of ["uniform", "quad"] for time_steps sampling `num_steps` method from `num_training_steps`.
  linear_start: float value for `beta` start value.
  linear_end: float value for `beta` end value.
  ddim_eta: float value for calculating `ddim_sigma`. 0 makes the sampling process deterministic.

Returns:
    A `StableDiffusion` instance.
"""

StableDiffusion.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__ + """
Model architectures:
  | Model               | Params | FLOPs   | Input               |
  | ------------------- | ------ | ------- | ------------------- |
  | ViTTextLargePatch14 | 123.1M | 6.67G   | [None, 77]          |
  | Encoder             | 34.16M | 559.6G  | [None, 512, 512, 3] |
  | UNet                | 859.5M | 404.4G  | [None, 64, 64, 4]   |
  | Decoder             | 49.49M | 1259.5G | [None, 64, 64, 4]   |
"""



import math
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format, initializers
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    add_with_layer_scale_and_drop_block,
    batchnorm_with_activation,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    inverted_residual_block,
    layer_norm,
    mlp_block,
    mhsa_with_multi_head_position,
    window_attention,
    MultiHeadPositionalEmbedding,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

LAYER_NORM_EPSILON = 1e-6

PRETRAINED_DICT = {
    "tiny_vit_11m": {"imagenet": {224: "6cc52dda567fd70f706d57c69dfd81d8"}, "imagenet21k-ft1k": {224: "f84673169e5c7a4ec526866da24cd5f4"}},
    "tiny_vit_21m": {
        "imagenet": {224: "5809a53bc3abe0785475c3f2d501a039"},
        "imagenet21k-ft1k": {224: "08d16dd06ddd85c2e4d2d15143c24607", 384: "fe6d364e99fa5a7f255ad3f3270bc962", 512: "ba5822042f0cb09bd8189290164b6ec3"},
    },
    "tiny_vit_5m": {"imagenet": {224: "a9c53f53e6da6a9b2edf6e773e5e402b"}, "imagenet21k-ft1k": {224: "cd10dbebf0645769dcda3224a0f330c4"}},
}


def tiny_vit_block(inputs, window_size=7, num_heads=4, mlp_ratio=4, layer_scale=0, drop_rate=0, activation="gelu", name=""):
    input_channel = inputs.shape[-1 if image_data_format() == "channels_last" else 1]

    """ attention """
    nn = inputs if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(inputs)  # channels_first -> channels_last
    nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, axis=-1, name=name + "attn_")
    nn = window_attention(
        nn, window_size, num_heads=num_heads, attention_block=mhsa_with_multi_head_position, use_bn=False, qkv_bias=True, out_bias=True, name=name + "attn_"
    )
    nn = nn if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(nn)  # channels_last, channels_first
    attn_out = add_with_layer_scale_and_drop_block(inputs, nn, layer_scale=layer_scale, drop_rate=drop_rate, name=name + "attn_")

    pre_mlp = depthwise_conv2d_no_bias(attn_out, kernel_size=3, strides=1, padding="same", name=name + "pre_mlp_")
    pre_mlp = batchnorm_with_activation(pre_mlp, activation=None, name=name + "pre_mlp_")

    """ MLP """
    nn = pre_mlp if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(pre_mlp)  # channels_first -> channels_last
    nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, axis=-1, name=name + "mlp_")
    nn = mlp_block(nn, input_channel * mlp_ratio, activation=activation, name=name + "mlp_")
    nn = nn if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(nn)  # channels_last, channels_first
    nn = add_with_layer_scale_and_drop_block(pre_mlp, nn, layer_scale=layer_scale, drop_rate=drop_rate, name=name + "mlp_")
    return nn


def TinyViT(
    num_blocks=[2, 2, 6, 2],
    out_channels=[64, 128, 160, 320],
    block_types=["conv", "transform", "transform", "transform"],
    num_heads=[2, 4, 5, 10],
    # window_sizes=[7, 7, 14, 7],
    window_ratios=[8, 4, 1, 1],  # For `input_shape=(224, 224, 3)` will be window_sizes=[7, 7, 14, 7], for `(384, 384, 3)` will be `[12, 12, 24, 12]`.
    mlp_ratio=4,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    dropout=0,
    layer_scale=0,
    classifier_activation="softmax",
    pretrained=None,
    model_name="tiny_vit",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)

    """ Stem """
    nn = conv2d_no_bias(inputs, out_channels[0] // 2, kernel_size=3, strides=2, padding="same", name="stem_1_")
    nn = batchnorm_with_activation(nn, activation=activation, name="stem_1_")
    nn = conv2d_no_bias(nn, out_channels[0], kernel_size=3, strides=2, padding="same", name="stem_2_")
    nn = batchnorm_with_activation(nn, activation=None, name="stem_2_")

    inverted_residual_block_kwargs = {
        "stride": 1,
        "expand": 4,
        "shortcut": True,
        "is_torch_mode": True,
        "use_last_bn_zero_gamma": True,
        "activation": activation,
    }

    """ stacks """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel, block_type) in enumerate(zip(num_blocks, out_channels, block_types)):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            name = stack_name + "downsample_"
            expand = out_channel / nn.shape[-1 if image_data_format() == "channels_last" else 1]
            nn = inverted_residual_block(nn, out_channel, stride=2, expand=expand, shortcut=False, is_torch_mode=True, activation=activation, name=name)

        is_conv_block = True if block_type[0].lower() == "c" else False
        num_head = num_heads[stack_id] if isinstance(num_heads, (list, tuple)) else num_heads
        # window_size = window_sizes[stack_id] if isinstance(window_sizes, (list, tuple)) else window_sizes
        window_ratio = window_ratios[stack_id] if isinstance(window_ratios, (list, tuple)) else window_ratios
        height, width = nn.shape[1:-1] if image_data_format() == "channels_last" else nn.shape[2:]
        window_size = [int(math.ceil(height / window_ratio)), int(math.ceil(width / window_ratio))]
        # print(f">>>> {window_size = }")
        for block_id in range(num_block):
            name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            if is_conv_block:
                nn = inverted_residual_block(nn, out_channel, **inverted_residual_block_kwargs, drop_rate=block_drop_rate, name=name)
                nn = activation_by_name(nn, activation=activation, name=name + "output_")
            else:
                nn = tiny_vit_block(nn, window_size, num_head, mlp_ratio, layer_scale, block_drop_rate, activation=activation, name=name)
            global_block_id += 1

    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn)
        nn = layer_norm(nn, epsilon=LAYER_NORM_EPSILON, name="pre_output_")
        if dropout > 0:
            nn = layers.Dropout(dropout, name="head_drop")(nn)
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)

    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "tinyvit", pretrained, MultiHeadPositionalEmbedding)
    return model


@register_model
def TinyViT_5M(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet21k-ft1k", **kwargs):
    return TinyViT(**locals(), model_name=kwargs.pop("model_name", "tiny_vit_5m"), **kwargs)


@register_model
def TinyViT_11M(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet21k-ft1k", **kwargs):
    out_channels = [64, 128, 256, 448]
    num_heads = [2, 4, 8, 14]
    return TinyViT(**locals(), model_name=kwargs.pop("model_name", "tiny_vit_11m"), **kwargs)


@register_model
def TinyViT_21M(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet21k-ft1k", **kwargs):
    out_channels = [96, 192, 384, 576]
    num_heads = [3, 6, 12, 18]
    return TinyViT(**locals(), model_name=kwargs.pop("model_name", "tiny_vit_21m"), **kwargs)



from keras_cv_attention_models.convnext.convnext import (
    global_response_normalize,
    ConvNeXt,
    ConvNeXtTiny,
    ConvNeXtSmall,
    ConvNeXtBase,
    ConvNeXtLarge,
    ConvNeXtXlarge,
    ConvNeXtXXlarge,
)
from keras_cv_attention_models.convnext.convnext_v2 import (
    ConvNeXtV2,
    ConvNeXtV2Atto,
    ConvNeXtV2Femto,
    ConvNeXtV2Pico,
    ConvNeXtV2Nano,
    ConvNeXtV2Tiny,
    ConvNeXtV2Base,
    ConvNeXtV2Large,
    ConvNeXtV2Huge,
)

__v1_head_doc__ = """
Keras implementation of [ConvNeXt](https://github.com/facebookresearch/ConvNeXt).
Paper [PDF 2201.03545 A ConvNet for the 2020s](https://arxiv.org/pdf/2201.03545.pdf).
"""

__common_head_doc__ = """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: output channels for each stack.
  stem_width: output dimension for stem block. Default `-1` means using `out_channels[0]`.
  model_name: string, model name.
"""

__common_tail_doc__ = """  layer_scale_init_value: layer scale init value, [Going deeper with Image Transformers](https://arxiv.org/abs/2103.17239).
      Default 1e-6 for v1, 0 for v2.
  use_grn: boolean value if use `global_response_normalize` block. Default False for v1, True for v2.
  head_init_scale: init head layer scale value if create model from scratch or fine-tune. Default `1`.
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `gelu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be a constant value like `0.2`,
      or a tuple value like `(0, 0.2)` indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch. or `0` to disable.
      Default 0.1.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `None`.
  dropout: dropout rate if top layers is included.
  pretrained: one of `None` (random initialization) or 'imagenet' (pre-training on ImageNet),
      or 'imagenet21k-ft1k' (pre-trined on ImageNet21k, fine-tuning on ImageNet),
      or 'clip-ft1k' (IN-1k fine-tune and IN-12k intermediate fine-tunes for CLIP models).
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.
"""

__v1_tail_doc__ = """
Model architectures:
  | Model                 | Params | FLOPs   | Input | Top1 Acc |
  | --------------------- | ------ | ------- | ----- | -------- |
  | ConvNeXtTiny          | 28M    | 4.49G   | 224   | 82.1     |
  | - ImageNet21k-ft1k    | 28M    | 4.49G   | 224   | 82.9     |
  | - ImageNet21k-ft1k    | 28M    | 13.19G  | 384   | 84.1     |
  | ConvNeXtSmall         | 50M    | 8.73G   | 224   | 83.1     |
  | - ImageNet21k-ft1k    | 50M    | 8.73G   | 224   | 84.6     |
  | - ImageNet21k-ft1k    | 50M    | 25.67G  | 384   | 85.8     |
  | ConvNeXtBase          | 89M    | 15.42G  | 224   | 83.8     |
  | ConvNeXtBase          | 89M    | 45.32G  | 384   | 85.1     |
  | - ImageNet21k-ft1k    | 89M    | 15.42G  | 224   | 85.8     |
  | - ImageNet21k-ft1k    | 89M    | 45.32G  | 384   | 86.8     |
  | ConvNeXtLarge         | 198M   | 34.46G  | 224   | 84.3     |
  | ConvNeXtLarge         | 198M   | 101.28G | 384   | 85.5     |
  | - ImageNet21k-ft1k    | 198M   | 34.46G  | 224   | 86.6     |
  | - ImageNet21k-ft1k    | 198M   | 101.28G | 384   | 87.5     |
  | ConvNeXtXLarge, 21k   | 350M   | 61.06G  | 224   | 87.0     |
  | ConvNeXtXLarge, 21k   | 350M   | 179.43G | 384   | 87.8     |
  | ConvNeXtXXLarge, clip | 846M   | 198.09G | 256   | 88.6     |
"""

ConvNeXt.__doc__ = __v1_head_doc__ + __common_head_doc__ + __common_tail_doc__ + __v1_tail_doc__

ConvNeXtTiny.__doc__ = __v1_head_doc__ + """
Args:
""" + __common_tail_doc__ + __v1_tail_doc__

ConvNeXtSmall.__doc__ = ConvNeXtTiny.__doc__
ConvNeXtBase.__doc__ = ConvNeXtTiny.__doc__
ConvNeXtLarge.__doc__ = ConvNeXtTiny.__doc__
ConvNeXtLarge.__doc__ = ConvNeXtTiny.__doc__
ConvNeXtXXlarge.__doc__ = ConvNeXtTiny.__doc__

__v2_head_doc__ = """
Keras implementation of [ConvNeXt](https://github.com/facebookresearch/ConvNeXt-V2).
Paper [PDF 2301.00808 ConvNeXt V2: Co-designing and Scaling ConvNets with Masked Autoencoders](https://arxiv.org/pdf/2301.00808.pdf).
"""

__v2_tail_doc__ = """
Model architectures:
  | Model              | Params | FLOPs  | Input | Top1 Acc |
  | ------------------ | ------ | ------ | ----- | -------- |
  | ConvNeXtV2Atto     | 3.7M   | 0.55G  | 224   | 76.7     |
  | ConvNeXtV2Femto    | 5.2M   | 0.78G  | 224   | 78.5     |
  | ConvNeXtV2Pico     | 9.1M   | 1.37G  | 224   | 80.3     |
  | ConvNeXtV2Nano     | 15.6M  | 2.45G  | 224   | 81.9     |
  | - ImageNet21k-ft1k | 15.6M  | 2.45G  | 224   | 82.1     |
  | - ImageNet21k-ft1k | 15.6M  | 7.21G  | 384   | 83.4     |
  | ConvNeXtV2Tiny     | 28.6M  | 4.47G  | 224   | 83.0     |
  | - ImageNet21k-ft1k | 28.6M  | 4.47G  | 224   | 83.9     |
  | - ImageNet21k-ft1k | 28.6M  | 13.1G  | 384   | 85.1     |
  | ConvNeXtV2Base     | 89M    | 15.4G  | 224   | 84.9     |
  | - ImageNet21k-ft1k | 89M    | 15.4G  | 224   | 86.8     |
  | - ImageNet21k-ft1k | 89M    | 45.2G  | 384   | 87.7     |
  | ConvNeXtV2Large    | 198M   | 34.4G  | 224   | 85.8     |
  | - ImageNet21k-ft1k | 198M   | 34.4G  | 224   | 87.3     |
  | - ImageNet21k-ft1k | 198M   | 101.1G | 384   | 88.2     |
  | ConvNeXtV2Huge     | 660M   | 115G   | 224   | 86.3     |
  | - ImageNet21k-ft1k | 660M   | 337.9G | 384   | 88.7     |
  | - ImageNet21k-ft1k | 660M   | 600.8G | 512   | 88.9     |
"""

ConvNeXtV2.__doc__ = __v2_head_doc__ + __common_head_doc__ + __common_tail_doc__ + __v2_tail_doc__

ConvNeXtV2Atto.__doc__ = __v2_head_doc__ + """
Args:
""" + __common_tail_doc__ + __v2_tail_doc__

ConvNeXtV2Femto.__doc__ = ConvNeXtV2Atto.__doc__
ConvNeXtV2Pico.__doc__ = ConvNeXtV2Atto.__doc__
ConvNeXtV2Nano.__doc__ = ConvNeXtV2Atto.__doc__
ConvNeXtV2Tiny.__doc__ = ConvNeXtV2Atto.__doc__
ConvNeXtV2Base.__doc__ = ConvNeXtV2Atto.__doc__
ConvNeXtV2Large.__doc__ = ConvNeXtV2Atto.__doc__
ConvNeXtV2Huge.__doc__ = ConvNeXtV2Atto.__doc__

global_response_normalize.__doc__ = __v2_head_doc__ + """
Global Response Normalize block

Args:
  inputs: input tensor.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 16, 256])
>>> nn = attention_layers.global_response_normalize(inputs)
>>> print(f"{nn.shape = }")
nn.shape = TensorShape([None, 14, 16, 256])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
>>> print({ii.name: ii.shape for ii in mm.weights})
{'channel_affine/weight:0': TensorShape([256]), 'channel_affine/bias:0': TensorShape([256])}
"""



from keras_cv_attention_models.llama2.llama2 import (
    LLaMA2,
    LLaMA2_15M,
    LLaMA2_42M,
    LLaMA2_110M,
    LLaMA2_7B,
    RunPrediction,
    PositionalEncodingFourierRot1D,
    RMSNorm,
    convert_huggingface_weights_to_h5,
)

__head_doc__ = """
Keras implementation of [Github facebookresearch/llama](https://github.com/facebookresearch/llama).
Paper [PDF 2307.09288 Llama 2: Open Foundation and Fine-Tuned Chat Models](https://arxiv.org/pdf/2307.09288.pdf).
"""

__tail_doc__ = """  vocab_size: model vocab size.
  max_block_size: number of tokens generated in each sample.
  include_top: boolena value if including output Dense head layer. Set false to exclude the head layer.
  dropout: float value for drop out rate for Embedding layer and attention blocks.
  activation: activation used in whole model, default `swish`.
  pretrained: None or "tiny_stories", or specific ".pt" or ".h5" file.
      - if "tiny_stories", will try to download and load ported weights if available.
      - if "xxx.pt", will try converting and loading weights from .pt file.
      - if "xxx.h5", will just load weights.
      - if None, will initialize model with ranbdom weights.

Returns:
    A `keras.Model` instance.
"""

LLaMA2.__doc__ = __head_doc__ + """
Args:
  num_blocks: num of `attention_fft_block`s.
  embedding_size: `attention_fft_block` block embedding size.
  hidden_divisible: int value making fft block hidden layer size multiple of large power of 2.
  num_heads: num of heads.
  block_use_bias: boolean value if using bias for `attention_fft_block` Dense layers.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model       | Params | FLOPs  | vocab_size | Val loss |
  | ----------- | ------ | ------ | ---------- | -------- |
  | LLaMA2_15M  | 24.41M | 4.06G  | 32000      | 1.072    |
  | LLaMA2_42M  | 58.17M | 50.7G  | 32000      | 0.847    |
  | LLaMA2_110M | 134.1M | 130.2G | 32000      | 0.760    |
  | LLaMA2_7B   | 6.74B  | 14.54T | 32000      |          |
"""

LLaMA2_15M.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

LLaMA2_42M.__doc__ = LLaMA2_15M.__doc__
LLaMA2_110M.__doc__ = LLaMA2_15M.__doc__
LLaMA2_7B.__doc__ = LLaMA2_15M.__doc__



from keras_cv_attention_models.vanillanet.vanillanet import (
    VanillaNet,
    VanillaNet5,
    VanillaNet6,
    VanillaNet7,
    VanillaNet8,
    VanillaNet9,
    VanillaNet10,
    VanillaNet11,
    VanillaNet12,
    VanillaNet13,
    set_leaky_relu_alpha,
    switch_to_deploy,
)
__head_doc__ = """
Keras implementation of [Github huawei-noah/VanillaNet](https://github.com/huawei-noah/VanillaNet).
Paper [PDF 2305.12972 VanillaNet: the Power of Minimalism in Deep Learning](https://arxiv.org/pdf/2305.12972.pdf).
"""

__tail_doc__ = """  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  deploy: boolean value if build a fused model. **Evaluation only, not good for training**.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `relu`.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `None`.
  pretrained: one of `None` (random initialization) or 'imagenet'.
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.

Train and Deploy:
- **Currently only works for Tensorflow**
- **`set_leaky_relu_alpha`** is introduced in the paper, used for changing activation alpha while training.
  Say `epochs=300, decay_epochs=100`, then `alpha = (current_epoch / decay_epochs) if current_epoch < decay_epochs else 1`.
- **`switch_to_deploy`** is used for converting a custom pretrained model to `deploy` version. The process is
  `fuse conv bn` -> `remove leaky_relu with alpha=1` -> `fuse 2 sequential conv`.
- Training process could be:
  >>> from keras_cv_attention_models import vanillanet
  >>> # Create a `deploy=False` for custom training
  >>> model = vanillanet.VanillaNet5(deploy=False, pretrained="imagenet")
  >>> print(f"{model.count_params() = }")
  >>> # model.count_params() = 22369712
  >>> # ...

  >>> # Run some training step with changing leaky_relu alpha
  >>> decay_epochs = 100
  >>> current_epoch = 10
  >>> model.set_leaky_relu_alpha(alpha=(current_epoch / decay_epochs) if current_epoch < decay_epochs else 1)
  >>> # Or set in actual callbacks
  >>> # on_epoch_end = lambda epoch, logs: model.set_leaky_relu_alpha(alpha=(epoch / decay_epochs) if epoch < decay_epochs else 1)
  >>> # act_learn = keras.callbacks.LambdaCallback(on_epoch_end=on_epoch_end)
  >>> # model.fit(..., callbacks=[..., act_learn])
  >>> # ...

  >>> # Fuse to deploy model
  >>> model.set_leaky_relu_alpha(alpha=1)  # This should already set when training reach `decay_epochs`
  >>> tt = model.switch_to_deploy()
  >>> deploy_save_name = model.name + "_deploy.h5"
  >>> tt.save(deploy_save_name)

  >>> # Further evaluation with deploy=True
  >>> bb = vanillanet.VanillaNet5(deploy=True, pretrained=deploy_save_name)
  >>> print(f"{bb.count_params() = }")
  >>> # bb.count_params() = 15523304
  >>> print(f"{np.allclose(model(tf.ones([1, 224, 224, 3])), bb(tf.ones([1, 224, 224, 3])), atol=1e-5) = }")
  >>> # np.allclose(model(tf.ones([1, 224, 224, 3])), bb(tf.ones([1, 224, 224, 3])), atol=1e-5) = True
"""

VanillaNet.__doc__ = __head_doc__ + """
Args:
  out_channels: output channels for each stack.
  strides: list value for strides value in each stack. Should be same length with `out_channels`.
  stem_width: stem output channels, default `-1` for using `out_channels[0]`.
  leaky_relu_alpha: float value in `[0, 1]` for initializing `LeakyReLU` alpha value, default `1.0` for identical, and `0` for `ReLU`.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model         | Params | FLOPs  | Input | Top1 Acc |
  | ------------- | ------ | ------ | ----- | -------- |
  | VanillaNet5   | 22.33M | 8.46G  | 224   | 72.49    |
  | - deploy=True | 15.52M | 5.17G  | 224   | 72.49    |
  | VanillaNet6   | 56.12M | 10.11G | 224   | 76.36    |
  | - deploy=True | 32.51M | 6.00G  | 224   | 76.36    |
  | VanillaNet7   | 56.67M | 11.84G | 224   | 77.98    |
  | - deploy=True | 32.80M | 6.90G  | 224   | 77.98    |
  | VanillaNet8   | 65.18M | 13.50G | 224   | 79.13    |
  | - deploy=True | 37.10M | 7.75G  | 224   | 79.13    |
  | VanillaNet9   | 73.68M | 15.17G | 224   | 79.87    |
  | - deploy=True | 41.40M | 8.59G  | 224   | 79.87    |
  | VanillaNet10  | 82.19M | 16.83G | 224   | 80.57    |
  | - deploy=True | 45.69M | 9.43G  | 224   | 80.57    |
  | VanillaNet11  | 90.69M | 18.49G | 224   | 81.08    |
  | - deploy=True | 50.00M | 10.27G | 224   | 81.08    |
  | VanillaNet12  | 99.20M | 20.16G | 224   | 81.55    |
  | - deploy=True | 54.29M | 11.11G | 224   | 81.55    |
  | VanillaNet13  | 107.7M | 21.82G | 224   | 82.05    |
  | - deploy=True | 58.59M | 11.96G | 224   | 82.05    |
"""

VanillaNet5.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

VanillaNet6.__doc__ = VanillaNet5.__doc__
VanillaNet7.__doc__ = VanillaNet5.__doc__
VanillaNet8.__doc__ = VanillaNet5.__doc__
VanillaNet9.__doc__ = VanillaNet5.__doc__
VanillaNet10.__doc__ = VanillaNet5.__doc__
VanillaNet11.__doc__ = VanillaNet5.__doc__
VanillaNet12.__doc__ = VanillaNet5.__doc__
VanillaNet13.__doc__ = VanillaNet5.__doc__



import os

is_tensorflow_backend, is_torch_backend, is_keras_core_backend = False, False, False
if "tensorflow" in os.getenv("KECAM_BACKEND", "tensorflow").lower():
    is_tensorflow_backend = True
elif "torch" in os.getenv("KECAM_BACKEND", "tensorflow").lower():
    is_torch_backend = True
else:
    is_keras_core_backend = True

if is_tensorflow_backend:
    try:
        import tensorflow as tf
    except ModuleNotFoundError as ee:
        print(">>>> [Warning] os environ 'KECAM_BACKEND' is 'tensorflow', but not installed.")
        try:
            import torch

            is_tensorflow_backend = False
            is_torch_backend = True
        except ModuleNotFoundError as ee:
            raise ModuleNotFoundError("Neither tensorflow nor torch found")


if is_tensorflow_backend:
    import tensorflow as tf
    from tensorflow.keras import layers, models, initializers, callbacks, metrics, losses, optimizers
    from tensorflow.keras.utils import register_keras_serializable, get_file
    from keras_cv_attention_models import tf_functional as functional
elif is_torch_backend:
    from keras_cv_attention_models.pytorch_backend import layers, models, functional, initializers, callbacks, metrics, losses, optimizers
    from keras_cv_attention_models.pytorch_backend.utils import register_keras_serializable, get_file

    print(">>>> Using PyTorch backend")
else:
    import keras_core
    from keras_core import layers, models, initializers, callbacks, metrics
    from keras_core.utils import register_keras_serializable, get_file

    # from keras_core import ops as functional
    from keras_cv_attention_models import keras_core_functional as functional


def backend():
    """backend() value can still be "torch" for `keras_core` backend, even `is_torch_backend` is False"""
    if is_tensorflow_backend:
        return tf.keras.backend.backend()
    elif is_torch_backend:
        return "torch"
    else:
        return keras_core.backend.backend()


def image_data_format():
    if is_tensorflow_backend:
        return tf.keras.backend.image_data_format()
    elif is_torch_backend:
        return "channels_first"
    else:
        return keras_core.backend.image_data_format()


__is_channels_last__ = image_data_format() == "channels_last"


def is_channels_last():
    return __is_channels_last__


def align_input_shape_by_image_data_format(input_shape):
    """Regard input_shape as force using original shape if len(input_shape) == 4,
    else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format,
    or if dynamic shape like `(None, None, 3)`, will regard the known shape 3 as channel dimension.

    Examples:
    >>> os.environ['KECAM_BACKEND'] = "torch"
    >>> from keras_cv_attention_models import backend
    >>> print(f"{backend.image_data_format() = }")
    >>> # backend.image_data_format() = 'channels_first'
    >>> print(backend.align_input_shape_by_image_data_format([224, 224, 3]))
    >>> # [3, 224, 224]
    >>> print(backend.align_input_shape_by_image_data_format([3, 224, 224]))
    >>> # [3, 224, 224]
    >>> print(backend.align_input_shape_by_image_data_format([None, None, 3]))
    >>> # [3, None, None]
    >>> print(backend.align_input_shape_by_image_data_format([None, 224, 224, 3]))
    >>> # [224, 224, 3]
    """
    # channel_axis = if image_data_format() == "channels_last" else
    if len(input_shape) == 4:  # Regard this as force using original shape
        return input_shape[1:]

    # Assume channel dimension is the one with min value in input_shape
    if None in input_shape:
        orign_channel_axis, channel_dim = min(enumerate(input_shape), key=lambda xx: xx[1] is None)
    else:
        orign_channel_axis, channel_dim = min(enumerate(input_shape), key=lambda xx: xx[1])
    orign_block_shape = [dim for axis, dim in enumerate(input_shape) if axis != orign_channel_axis]
    aligned = [*orign_block_shape, channel_dim] if image_data_format() == "channels_last" else [channel_dim, *orign_block_shape]
    if aligned[1] != input_shape[1] or aligned[-1] != input_shape[-1]:
        print(">>>> Aligned input_shape:", aligned)
    return aligned


def in_train_phase(train_phase, eval_phase, training=None):
    if is_tensorflow_backend:
        return tf.keras.backend.in_train_phase(train_phase, eval_phase, training=training)
    elif is_torch_backend:
        return train_phase if training else eval_phase  # [TODO]
    else:
        return train_phase if training else eval_phase  # [TODO]


def numpy_image_resize(inputs, target_shape, method="bilinear", antialias=False, is_source_channels_last=True):
    import numpy as np

    ndims = len(inputs.shape)
    if ndims < 2 or ndims > 4:
        raise ValueError("inputs with shape={}, ndims={} not supported, ndims must in [2, 4]".format(ipnuts.shape, ndims))

    target_shape = [int(ii) for ii in target_shape]
    if is_source_channels_last:
        inputs = inputs if ndims == 4 else (inputs[None] if ndims == 3 else inputs[None, :, :, None])
        inputs = inputs if image_data_format() == "channels_last" else inputs.transpose([0, 3, 1, 2])

        inputs = functional.resize(inputs, target_shape, method=method, antialias=antialias)
        inputs = inputs.detach() if hasattr(inputs, "detach") else inputs
        inputs = inputs.numpy() if hasattr(inputs, "numpy") else np.array(inputs)

        inputs = inputs if image_data_format() == "channels_last" else inputs.transpose([0, 2, 3, 1])
        inputs = inputs if ndims == 4 else (inputs[0] if ndims == 3 else inputs[0, :, :, 0])
    else:
        inputs = inputs if ndims == 4 else (inputs[None] if ndims == 3 else inputs[None, None, :, :])
        inputs = inputs.transpose([0, 2, 3, 1]) if image_data_format() == "channels_last" else inputs

        inputs = functional.resize(inputs, target_shape, method=method, antialias=antialias)
        inputs = inputs.detach() if hasattr(inputs, "detach") else inputs
        inputs = inputs.numpy() if hasattr(inputs, "numpy") else np.array(inputs)

        inputs = inputs.transpose([0, 3, 1, 2]) if image_data_format() == "channels_last" else inputs
        inputs = inputs if ndims == 4 else (inputs[0] if ndims == 3 else inputs[0, 0])
    return inputs
